import pandas as pd
from tqdm import tqdm

tqdm.pandas()
from humobi.measures.individual import jump_lengths
from humobi.preprocessing.filters import next_location_sequence
from humobi.tools.processing import convert_to_distribution


def dist_travelling_distance(trajectories_frame, bin_size=None, n_bins=20):
	"""
	Calculates the distribution of travelling distances for each user

	Args:
		trajectories_frame: TrajectoriesFrame class object
		bin_size (default = None): size of a bin for histogram (if used, num_of_classes cannot be determined)
		n_bins (default = 20): number of groups in histogram (if used, bin_size cannot be determined)

	Returns:
		Traveling distances for each user
	"""
	jumps = jump_lengths(trajectories_frame).dropna().droplevel(1)
	jumps = jumps[jumps != 0]
	jumps = convert_to_distribution(jumps, bin_size=bin_size, num_of_classes=n_bins)
	return jumps


def flows(trajectories_frame, flows_type='all'):
	"""
	Calculates the number of flows for each aggregation cell. All flows, only incoming or only outgoing flows can be
	counted.

	Args:
		trajectories_frame: TrajectoriesFrame class object
		flows_type: Type of flows to be counted (possible values: 'all', 'incoming', 'outgoing') (default = 'all')

	Returns:
		a DataFrame with flows grouped by cells of aggregation grid
	"""
	trajectories_frame = trajectories_frame.dropna()
	next_locations = next_location_sequence(trajectories_frame)
	next_locations['geometry'] = next_locations['geometry'].astype(str)
	double_flows = next_locations.groupby(level=0).progress_apply(lambda x: x[1:-1].groupby('geometry').count()). \
		droplevel(0)
	if flows_type == 'all':
		double_flows = double_flows * 2
	if flows_type == 'incoming':
		trajectories_edges = [-1]
	elif flows_type == 'outgoing':
		trajectories_edges = [0]
	else:
		trajectories_edges = [0, -1]
	single_flows = next_locations.groupby(level=0).progress_apply(
		lambda x: x.iloc[trajectories_edges].groupby('geometry').count()). \
		droplevel(0)
	total_flows = pd.concat([double_flows, single_flows])
	total_flows = total_flows[total_flows.columns[0]]
	return total_flows.groupby('geometry').sum().sort_values(ascending=False)



import pandas as pd
import geopandas as gpd
import sys

sys.path.append("..")
from humobi.tools.processing import normalize
from humobi.models.spatial_tools.misc import rank_freq


def calculate_distances(gs1, gs2):
	"""
	Calculates the distance (ellipsoidal) between to GeoSeries
	:param gs1: GeoSeries1
	:param gs2: GeoSeries2
	:return: The GeoSeries of distances
	"""
	return gs1.distance(gs2)


def commute_distances(trajectories_frame, quantity=2):
	"""
	Calculates the commuting distances for each user. Quantity regulates the number of locations to which the distance
	is calculated.
	:param trajectories_frame: TrajectoriesFrame class object
	:param quantity: The number of locations to which the distance will be calculated
	:return: The DataFrame of distances to the locations
	"""
	sig_frame = rank_freq(trajectories_frame, quantity=quantity)
	indices = []
	distances = {}
	for n in range(quantity - 1):
		for k in range(n + 1, quantity):
			indices.append((n, k))
	indices = sorted(indices, key=lambda x: x[1])
	prev_k = 1
	df_list = []
	for n, k in indices:
		if k != prev_k:
			distances[prev_k] = pd.concat(df_list, axis=1)
			df_list = []
		prev_k = k
		gs1 = gpd.GeoDataFrame(sig_frame.iloc[:, n], geometry=n)
		gs2 = gpd.GeoDataFrame(sig_frame.iloc[:, k], geometry=k)
		distance = calculate_distances(gs1, gs2)
		df_list.append(pd.DataFrame([distance, sig_frame.iloc[:, n]]).T)  # distance and starting point geometry
	distances[k] = pd.concat(df_list, axis=1)
	return distances


def commute_distances_to_2d_distribution(commute_distances_frame, layer, crs=None, return_centroids=False):
	"""
	Converts commuting distances into 2D distribution of median commuting distances.
	:param commute_distances_frame: DataFrame with commuting distances to important locations
	:param layer: Aggregation layer
	:param crs: CRS of output data
	:param return_centroids: Whether full layer or centroids should be returned
	:return: 2D distribution of median commuting distances
	"""
	commute_distributions = {}
	for k,v in commute_distances_frame.items():
		commute_distances_frame = gpd.GeoDataFrame(v.reset_index(drop=True), geometry=0)
		if crs:
			pass
		else:
			crs = layer.crs
		grouped = gpd.tools.sjoin(commute_distances_frame, layer, how='right').groupby(level=0)
		to_concat = []
		for uid, vals in grouped:
			if vals.iloc[:, 1:k+1].isna().all()[0]:
				pass
			else:
				vals.iloc[0, 1:k+1] = vals.iloc[:, 1:k+1].median()
			to_concat.append(vals.iloc[0, :])
		merged = pd.concat(to_concat, axis=1).T
		merged = merged.rename(columns={"Unnamed 0": 'distance'})
		merged = gpd.GeoDataFrame(merged, geometry='geometry').fillna(0)
		if return_centroids:
			merged['geometry'] = merged['geometry'].centroid
		merged = merged[['distance','geometry']]
		commute_distributions[k] = merged
	return commute_distributions


def convert_to_2d_distribution(significant_places_frame, layer, quantity=1, crs=None, return_centroids=False):
	"""
	A general function for converting points into 2D distribution
	:param significant_places_frame: Points layer - usually significant places
	:param layer: Aggregation layer
	:param quantity: The number of layers to return - if there is than one point per index, more layers can be returned
	:param crs: CRS of output data
	:param return_centroids: Whether full layer or centroids should be returned
	:return: List of DataFrames with 2D distributions
	"""
	distributions = []
	significant_places_frame = significant_places_frame.reset_index()
	if crs:
		pass
	else:
		crs = layer.crs
	if isinstance(significant_places_frame, pd.DataFrame):
		for to_convert in range(quantity):
			col_name = significant_places_frame.columns[to_convert + 1]
			sig_locs = significant_places_frame.loc[:, col_name]
			sig_locs = sig_locs.dropna()
			if sig_locs.dtype == 'geometry':
				sig_locs_data = gpd.GeoDataFrame(sig_locs, crs=crs, geometry=col_name)
			elif sig_locs.dtype == 'str':
				sig_locs = sig_locs.str.split(";", expand=True, ).astype(float)
				sig_locs_data = gpd.GeoDataFrame(significant_places_frame,
				                                 geometry=gpd.points_from_xy(*[sig_locs[x] for x in sig_locs]), crs=crs)
			account = gpd.tools.sjoin(sig_locs_data, layer).groupby('index_right').count().iloc[:, 0]
			layer_counted = pd.merge(layer, account, left_index=True, right_index=True, how='outer').fillna(0)
			layer_counted = layer_counted.rename(columns={to_convert: 'counted'})
			layer_counted = normalize(layer_counted, 'counted')
			if return_centroids:
				layer_counted['geometry'] = layer_counted['geometry'].centroid
			distributions.append(layer_counted)
	return distributions


import pandas as pd, geopandas as gpd
import sys
sys.path.append("..")
from humobi.models.temporal_tools import when
from humobi.models.spatial_modules import where
# from humobi.misc import export_to_file
WEIGHT = False


def data_generator(start, agents, clusters, circadian_rhythm, end=None, duration=None, output_format = 'csv', output_style = 'user'):
	"""
	Produces a time-step-based simulation for every agent.

	Args:
		start: contains a string value which determines a start of simulation, format: 'day.month.year'
		agents: contains a list with instances of Agent Class
		clusters: contains a list with clusters numbers
		end: contains a string value which determines an end of simulation, format: 'day.month.year'
		duration: (an optional parameter) contains a duration of simulation
		output_format: contains a string with the format of output file: csv,feather or parquet
		output_style: contains a string 'user' or 'slot' which determines a layout of the output file

	Returns:
		a list with agents for each slot of simulation
	"""
	start = pd.to_datetime(start)
	if duration is None:
		if end is None:
			raise ValueError("You have to declare the end value")
		end = pd.to_datetime(end)
		duration = end-start
	else:
		duration = pd.Timedelta('{} days'.format(duration))
		end = start + duration
	slots_amount = 24 * duration.days
	time_slots = pd.date_range(start, end, slots_amount + 1)
	for slot in time_slots:
		for cluster in clusters:
			users = [agent for agent in agents if agent.cluster == cluster]
			for user in users:
				when.when(user, slot, circadian_rhythm)
				where.where(user)
	# export_to_file.export_to_file(agents, time_slots, output_format, output_style)


import sys
sys.path.append("..")
from humobi.misc import create_grid
from humobi.structures.trajectory import TrajectoriesFrame
from humobi.models.temporal_tools import cluster_traj
from humobi.models.spatial_tools import misc, distributions, generating, filtering
from humobi.models.spatial_tools.misc import rank_freq
from humobi.models.agent_module.generate_agents import generate_agents
from humobi.preprocessing.spatial_aggregation import LayerAggregator
from math import ceil


def data_sampler(input_data, aggregation, weigthed):
    """
	:param input_data: 
	:return: 
	"""
    if input_data.crs.coordinate_system.name == 'ellipsoidal':
        input_data = input_data.to_crs("epsg:3857")
    if isinstance(aggregation, int):
        layer = create_grid.create_grid(input_data, resolution=aggregation)
    elif isinstance(aggregation, str):
        layer = LayerAggregator(aggregation)
        if not layer.layer.crs == input_data.crs:
            layer.layer = layer.layer.to_crs(input_data.crs)
    layer = filtering.filter_layer(layer, input_data)
    circadian_collection, cluster_association, cluster_share = cluster_traj.cluster_trajectories(input_data,
                                                                                                 weights=weigthed,
                                                                                                 quantity=2,
                                                                                                 aux_cols=['a_t2m',
                                                                                                           'a_wind_speed',
                                                                                                           'a_wind_dir',
                                                                                                           'a_tcc'])  # 'a_t2m','a_wind_speed','a_wind_dir','a_tcc'
    commute_dist = distributions.commute_distances(input_data, quantity=2)
    unique_labels = set(cluster_association.values()).difference(set([-1]))
    sig_frame = rank_freq(input_data, quantity=2)
    cluster_spatial_distributions = {}
    cluster_commute_distributions = {}
    for n in [unique_labels][0]:
        group_indicies = [k for k, v in cluster_association.items() if v == n]
        group_sig_frame = sig_frame.loc[group_indicies]
        group_commute_dist = {k: v.loc[group_indicies] for k, v in commute_dist.items()}
        dist_list = distributions.convert_to_2d_distribution(group_sig_frame, layer, return_centroids=True, quantity=2)
        commute_distributions = distributions.commute_distances_to_2d_distribution(group_commute_dist, layer,
                                                                                   return_centroids=True)
        cluster_spatial_distributions[n] = dist_list
        cluster_commute_distributions[n] = commute_distributions
    to_generate = 217
    generated_agents = []
    for label, share in cluster_share.items():
        amount = ceil(share * to_generate)
        current_spatial_distributions = cluster_spatial_distributions[label]
        current_commute_distributions = cluster_commute_distributions[label]
        home_positions = generating.generate_points_from_distribution(current_spatial_distributions[0], amount)
        work_positions = generating.select_points_with_commuting(home_positions, current_spatial_distributions[1],
                                                                 current_commute_distributions, spread=.05)
        activity_areas = generating.generate_activity_areas('ellipse', home_positions, work_positions, layer, 1.0)
        agents = generate_agents(amount, label, home_positions, work_positions, activity_areas)
        generated_agents += agents
    circadian_rhythm = cluster_traj.circadian_rhythm_extraction(circadian_collection)




class Agent:
	"""
	An Agent class.

	Args:
		sig_locs: contains significant locations,
		temporal_mechanism: contains temporal mechanism,
		spatial_mechanism: contains spatial mechanism,
		cluster: contains the label of cluster,
		activity_area: contains the area of activity.
	"""

	def __init__(self, sig_locs, temporal_mechanism, spatial_mechanism, cluster, activity_area):
		"""
		Initialises and agent and assigns him temporal and spatial mechanisms of movement.
		"""
		self._sig_locs = sig_locs
		self._pref_return = {}
		self._history = []
		self._temporal_mechanism = temporal_mechanism
		self._spatial_mechanism = spatial_mechanism
		self._current_loc = None
		self._cluster = cluster
		self._activity_area = activity_area

	def __str__(self):
		return 'Current location: ' + str(self.current_loc) + ' Home position: ' + str(
			self.sig_locs[0]) + ' Cluster: ' + str(self.cluster)

	def __repr__(self):
		return 'Current location: ' + str(self.current_loc) + ' Home position: ' + str(
			self.sig_locs[0]) + ' Cluster: ' + str(self.cluster)

	@property
	def sig_locs(self):
		return self._sig_locs

	@sig_locs.setter
	def sig_locs(self, new_value):
		self._sig_locs = new_value

	@property
	def pref_return(self):
		return self._pref_return

	@pref_return.setter
	def pref_return(self, new_dict):
		self._pref_return = new_dict

	@property
	def history(self):
		return self._history

	@history.setter
	def history(self, new_list):
		self._history = new_list

	@property
	def temporal_mechanism(self):
		return self._temporal_mechanism

	@temporal_mechanism.setter
	def temporal_mechanism(self, new_value):
		self._temporal_mechanism = new_value

	@property
	def spatial_mechanism(self):
		return self._spatial_mechanism

	@spatial_mechanism.setter
	def spatial_mechanism(self, new_value):
		self._spatial_mechanism = new_value

	@property
	def current_loc(self):
		return self._current_loc

	@current_loc.setter
	def current_loc(self, new_value):
		self._current_loc = new_value

	@property
	def cluster(self):
		return self._cluster

	@cluster.setter
	def cluster(self, new_value):
		self._cluster = new_value

	@property
	def activity_area(self):
		return self._activity_area

	@activity_area.setter
	def activity_area(self, new_value):
		self._activity_area = new_value



import pandas as pd
from humobi.structures.trajectory import TrajectoriesFrame
from humobi.preprocessing.filters import fill_gaps
import concurrent.futures as cf
from tqdm import tqdm


class TemporalAggregator():
	"""
	A class for temporal aggregation of TrajectoriesFrame. It aggregates each users' trajectory separately and returns
	single TrajectoriesFrame.
	"""

	def __init__(self, resolution):
		"""
		Class initialisation

		Args:
			resolution: resolution to which data will be aggregated
		"""
		self._resolution = resolution

	@property
	def resolution(self):
		return self._resolution

	def _user_aggregate(self, uid, single_user):
		"""
		Single trajectory aggregation.  For each time frame the longest visited location is selected.

		Args:
			uid: user's identifier
			single_user: user's trajectory

		Returns:
			An aggregated trajectory
		"""
		single_user['temp'] = range(len(single_user))
		df_time = (single_user[['temp', 'start', 'end']].set_index('temp').stack().reset_index(level=-1, drop=True).rename('time').to_frame())
		df_time['time'] = pd.to_datetime(df_time['time'])
		df_time = (df_time.groupby('temp').apply(lambda x: x.set_index('time').resample(self.resolution).asfreq()).reset_index())
		df_time = df_time.merge(single_user[single_user.columns.values])
		df_time['start'] = pd.to_datetime(df_time['start'])
		df_time['end'] = pd.to_datetime(df_time['end'])
		df_time = df_time.set_index('time')
		grouped = df_time.resample(self.resolution)
		for_concat = []
		tdelta = pd.Timedelta(self.resolution)
		for time_bin, data_bin in grouped:
			if len(data_bin) > 1:
				start = data_bin.index.min()
				end = start + tdelta
				longest_stay = data_bin.groupby('temp').apply(
					lambda x: x['end'] - x['start'] if ((x['end'] < end).all() and (x['start'] > start).all()).all() else end - x[
						'start'] if (x['end'] > end).all() else x['start'] - start).idxmax()[0]
				selected = data_bin[data_bin['temp'] == longest_stay]
			else:
				selected = data_bin
			for_concat.append(selected)
		df_time = pd.concat(for_concat)
		df_time = pd.concat({uid: df_time})
		return df_time

	def aggregate(self, trajectories_frame, fill_method = None, drop_empty = False, parallel = True):
		"""
		Aggregates TrajectoriesFrame temporally to given temporal resolution.
		For each time frame the longest visited location is selected.

		Args:
			trajectories_frame: TrajectoriesFrame class object
			fill_method: Empty records filling method, if 'None' then data is not filled in. Any other value causes use of forward fill.
			drop_empty (default = False): Determines if empty records should be removed.
			parallel (default = True): Determines if parallel computing should be used.

		Returns:
			A temporally aggregated TrajectoriesFrame object
		"""
		if not hasattr(trajectories_frame, '_geom_cols'):
			trajectories_frame = TrajectoriesFrame(trajectories_frame)
		trajectories_frame_grouped = trajectories_frame.groupby(level=0)
		for_concat = []
		if parallel:
			with cf.ThreadPoolExecutor() as executor:
				indis = [indi for indi, val in trajectories_frame_grouped]
				vals = [val for indi, val in trajectories_frame_grouped]
				results = list(tqdm(executor.map(self._user_aggregate, indis, vals), total=len(trajectories_frame_grouped)))
			for result in results:
				for_concat.append(result)
		else:
			for uid, user_data in trajectories_frame_grouped:
				if not (user_data['lat'].isna()).all():
					single_user_aggregated = self._user_aggregate(uid, user_data)
					for_concat.append(single_user_aggregated)
		trajectories_frame = pd.concat(for_concat).groupby(level=0).apply(lambda x: x.groupby(pd.Grouper(level=1,freq=self.resolution)).agg('first'))
		if fill_method is not None:
			trajectories_frame = fill_gaps(trajectories_frame, fill_method)
			trajectories_frame = fill_gaps(trajectories_frame,'bfill')
		trajectories_frame = trajectories_frame.drop('geometry', axis=1)
		if drop_empty:
			trajectories_frame = trajectories_frame[~trajectories_frame['lat'].isna()]

		trajectories_frame = TrajectoriesFrame(trajectories_frame)
		return trajectories_frame



import pandas as pd
import geopandas as gpd
import numpy as np
import re
import geofeather as gf
import warnings


def infer_datetime(trajectories_frame):
	"""
	This function infers the timestamp column index.

	Args:
		trajectories_frame (TrajectoriesFrame): TrajectoriesFrame class object

	Returns:
		date column index and time column index or datetime column index
	"""
	single_row = trajectories_frame.head(1)
	date_col = None
	time_col = None
	for name, val in single_row.iteritems():
		val = val.iloc[0]
		if re.match(r".+(\/).+(\/).+", str(val)):
			date_col = name
		elif re.match(r".+(-).+(-).+", str(val)):
			date_col = name
		if re.match(r".{1,2}(:).{1,2}(:).{1,2}", str(val)):
			time_col = name

	if date_col == time_col:
		return date_col
	elif isinstance(date_col, str) or isinstance(date_col, int) and time_col is None:
		return date_col
	else:
		return date_col, time_col


def infer_geometry(trajectories_frame):
	"""
	This function infers the index of column with geometry
	
	Args:
		trajectories_frame (TrajectoriesFrame): TrajectoriesFrame class object
	
	Returns:
		geometry column index
	"""
	latitude_column = None
	longitude_column = None
	if trajectories_frame.columns.dtype is 'str' or trajectories_frame.columns.is_object():
		try:
			latitude_column = [x for x in trajectories_frame.columns if 'lat' in x.lower()][0]
		except:
			pass
		try:
			longitude_column = [x for x in trajectories_frame.columns if 'lon' in x.lower()][0]
		except:
			pass
	if not latitude_column or not longitude_column:
		try:
			latitude_column, longitude_column = trajectories_frame.columns[
				trajectories_frame.dtypes == 'float64'].values
			warnings.warn("TrajectoriesFrame: Selecting first two columns of float")
		except:
			trajectories_frame['lat'] = 0
			trajectories_frame['lon'] = 0
			latitude_column = 'lat'
			longitude_column = 'lon'
			warnings.warn("TrajectoriesFrame: No coordinates, adding empty floats")
	return latitude_column, longitude_column


def infer_id(trajectories_frame):
	"""
	This function infers the index of column with user identifier

	Args:
		trajectories_frame (TrajectoriesFrame): TrajectoriesFrame class object

	Returns:
		identifier column index
	"""
	if trajectories_frame.columns.dtype is 'str' or trajectories_frame.columns.is_object():
		try:
			id_col = [x for x in trajectories_frame.columns if 'id' in x.lower()][0]
			return id_col
		except:
			samples = trajectories_frame.sample(1000,replace=True)
			id_col = np.argmax([samples.groupby(x).size().mean() for x in samples.columns])
			id_col = trajectories_frame.columns[id_col]
			warnings.warn("ID column was infered but to avoid mistakes ensure that TrajectoriesFrame contains an ID "
			              "column (with id in the name) and reread the frame!")
			return id_col


class TrajectoriesFrame(gpd.GeoDataFrame):
	"""
	This is a basic class of TrajectoriesFrame. It stores user traces with MultiIndex - sorted by ID, and by timestamp.
	This class inherits from GeoPandas DataFrame, therefore has geographical properties.
	"""

	_metadata = ['_crs', '_geom_cols', ]

	@property
	def _constructor(self):
		def _c(data):
			return TrajectoriesFrame(data).__finalize__(self)

		return _c

	def __init__(self, data=None, params={}):
		"""
		Creates TrajectoriesFrame from input - can be DataFrame or file path

		Args:
			data: DataFrame or file path
			params: dictionary of parameters accepted by DataFrame class
		"""
		try:
			self._crs = data.crs
		except AttributeError:
			passed_crs = params.pop('crs', None)
			self._crs = passed_crs

		if isinstance(data, str):
			data = pd.read_csv(data, **params)

		columns_user_defined = params.pop('columns', None)
		self._geom_cols = params.pop('geom_cols', None)

		if isinstance(data, pd.DataFrame):
			if not data.index.nlevels > 1:
				if len(data.select_dtypes(include=[np.datetime64])) > 1:
					if 'datetime' in data.columns:
						date_cols = 'datetime'
					elif 'time' in data.columns:
						date_cols = 'time'
					else:
						date_cols = infer_datetime(data)
					if isinstance(date_cols, str) or isinstance(date_cols, int):
						data["datetime"] = pd.to_datetime(data[date_cols])
					else:
						data["datetime"] = pd.to_datetime(data[date_cols[0]] + " " + data[date_cols[1]])
					if date_cols != 'datetime':
						data = data.drop(date_cols, axis=1)
				id_col = infer_id(data)
				if id_col != 'user_id':
					data['user_id'] = data[id_col]
					data = data.drop(id_col,axis=1)
				data = data.set_index(['user_id', "datetime"])
			if self._geom_cols is None:
				self._geom_cols = infer_geometry(data)

			data = gpd.GeoDataFrame(data, geometry=gpd.points_from_xy(*[data[x] for x in self._geom_cols]),
			                        crs=self._crs)

		super().__init__(data)

		if columns_user_defined:
			data.columns = columns_user_defined

		if isinstance(data, gpd.GeoDataFrame):
			if self._crs is None:
				self.crs = passed_crs
			self.geom_cols = self._geom_cols

	@property
	def _base_class_view(self):
		"""
		Forces DataFrame view in debugging mode for some IDEs.
		"""
		return pd.DataFrame(self)

	def to_geofeather(self, path):
		"""
		Writes TarjectoriesFrame to geofeather file

		Args:
			path: write path

		Returns:
			file path
		"""
		gf.to_geofeather(self.reset_index(), path)
		return path

	@classmethod
	def from_geofeather(self, path):
		"""
		Reads from geofeather file

		Args:
			path: file path

		Returns:
			TrajectoriesFrame object
		"""
		return TrajectoriesFrame(gf.from_geofeather(path))

	def to_shapefile(self, path):
		"""
		Writes TrajectoriesFrame to shapefile

		Args:
			path: write path

		Returns:
			file path
		"""
		warnings.warn("Saving to shapefile forces conversion of datetime into string")
		self = self.reset_index()
		self['datetime'] = str(self['datetime'])
		self.to_file(path)
		return path

	def uloc(self, uid):
		"""
		Returns a part of TrajectoriesFrame for user ID

		Args:
			uid: user id

		Returns:
			TrajectoriesFrame of selected users
		"""
		return self.loc[uid]

	def get_users(self):
		"""
		Returns all the users' IDs

		Returns:
			Array of all users' identifiers
		"""
		return self.index.get_level_values(0).unique().values

	def to_crs(self, dest_crs, cur_crs=None):
		"""
		Transforms geometry of TrajectoryFrame into different Coordinate Reference System.

		Args:
			dest_crs: CRS to which TrajectoriesFrame will be converted
			cur_crs (default=None): Current CRS of a TrajectoryFrame. If not determined, CRS will be read from the metadata.

		Returns:
			Converted TrajectoryFrame. This also alters the instance itself.
		"""
		if cur_crs is None:
			cur_crs = self.crs
		self = super(TrajectoriesFrame, self).to_crs(dest_crs)
		self[[x for x in self._geom_cols]] = pd.concat([self.geometry.x, self.geometry.y], axis=1)
		return self



# IMPORTS
import numpy as np
import pandas as pd
import geopandas as gpd
from humobi.structures.trajectory import TrajectoriesFrame
from tqdm import tqdm
tqdm.pandas()
import concurrent.futures as cf
from itertools import repeat


def fill_gaps(trajectories_frame, fill_method):
	"""
	Filling empty TrajectoriesFrame records

	Args:
		trajectories_frame: TrajectoriesFrame class object
		fill_method: how to fill the data, 'ffill' - forward fill, 'bfill" - backward fill

	Returns:
		Filled TrajectoriesFrame object
	"""
	# TODO: Add more filling routines
	while trajectories_frame['geometry'].isna().any():
		if fill_method == 'ffill':
			trajectories_frame = trajectories_frame.fillna(trajectories_frame.shift())
		if fill_method == 'bfill':
			trajectories_frame = trajectories_frame.fillna(trajectories_frame.shift(-1))
	return trajectories_frame


def next_location_sequence(trajectories_frame):
	"""
	Converts TrajectoriesFrame to the next location sequence

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		converted TrajectoriesFrame
	"""
	trajectories_frame = trajectories_frame[~trajectories_frame.isna().iloc[:,0]]
	crs = trajectories_frame.crs
	try:
		geom_cols = trajectories_frame.geom_cols
	except:
		geom_cols = ['lat', 'lon']
	coordinates_frame = trajectories_frame[[geom_cols[0], geom_cols[1]]]
	trajectories_frame = TrajectoriesFrame(
		trajectories_frame[(coordinates_frame != coordinates_frame.shift()).any(axis=1)],
		{'crs': crs, 'geom_cols': geom_cols})
	return trajectories_frame


def _user_stops(indi, single_trajectory, distance_condition, time_condition):
	"""
	Detects stops in a single user's trajectory

	Args:
		indi: user identifier
		single_trajectory: user's trajectory
		distance_condition: distance threshold for stop detection
		time_condition: time threshold for stop detection

	Returns:
		TrajectoriesFrame with records indicated as stops in 'is_stop' column
	"""
	if 'datetime' not in single_trajectory.columns:
		single_trajectory = single_trajectory.reset_index()
	single_trajectory['datetime'] = pd.to_datetime(single_trajectory['datetime'])
	if single_trajectory.crs is None:
		single_trajectory.crs = "EPSG:4326"
	single_trajectory = gpd.GeoDataFrame(single_trajectory, crs=single_trajectory.crs)
	single_trajectory.geometry.crs = single_trajectory.crs
	distances = np.array(single_trajectory.to_crs("EPSG:3857"))
	starting_index = 0
	stops = []
	j = 0
	geom_loc = single_trajectory.columns.get_loc('geometry')
	time_loc = single_trajectory.columns.get_loc('datetime')
	while starting_index + j < len(distances) - 1:
		j += 1
		ending_index = starting_index + j
		actual_distance = distances[starting_index][geom_loc].distance(distances[ending_index][geom_loc])
		if actual_distance > distance_condition:
			if ending_index - starting_index > 1:
				if time_condition:
					if distances[ending_index-1][time_loc] - distances[starting_index][time_loc] > pd.Timedelta(time_condition):
						pass
					else:
						starting_index = ending_index
						j = 0
						continue
				stops.append(list(range(starting_index, ending_index)))
			starting_index = ending_index
			j = 0
	single_trajectory['is_stop'] = False
	for stop in stops:
		selected = single_trajectory.iloc[stop]
		mean_lat = selected['lat'].mean()
		mean_lon = selected['lon'].mean()
		single_trajectory.iloc[stop, single_trajectory.columns.get_loc('lat')] = mean_lat
		single_trajectory.iloc[stop, single_trajectory.columns.get_loc('lon')] = mean_lon
		single_trajectory.iloc[stop, single_trajectory.columns.get_loc('is_stop')] = True
	return indi, single_trajectory


def stop_detection(trajectories_frame, distance_condition=300, time_condition='10 min'):
	"""
	Detects all stops in the TrajectoriesFrame. Uses multithreading.

	Args:
		trajectories_frame: TrajectoriesFrame class object
		distance_condition: distance threshold for stop detection
		time_condition: time threshold for stop detection

	Returns:
		TrajectoriesFrame with records indicated as stops in 'is_stop' column
	"""
	result_dic = {}
	with cf.ThreadPoolExecutor() as executor:
		args = [val for indi, val in trajectories_frame.groupby(level=0)]
		ids = [indi for indi, val in trajectories_frame.groupby(level=0)]
		results = list(tqdm(executor.map(_user_stops, ids, args, repeat(distance_condition), repeat(time_condition)),
		                    total=len(ids)))
	for result in results:
		result_dic[result[0]] = result[1]
	detected = pd.concat([x for x in result_dic.values()])
	detected = TrajectoriesFrame(detected)
	return detected



import sys
sys.path.append("..")
import pandas as pd
from tqdm import tqdm
tqdm.pandas()


def fraction_of_empty_records(trajectories_frame, resolution):
	"""
	Calculates the fraction q of empty records per user in a TrajectoriesFrame at given temporal resolution

	Args:
		trajectories_frame: TrajectoriesFrame class object
		resolution: time unit at which the fraction of empty records will be determined

	Returns:
		A pandas Series with fractions of empty record for each user
	"""
	resampled = trajectories_frame.groupby(level=0).resample(resolution, level=1).count().iloc[:,0]
	fractions = resampled.groupby(level=0).progress_apply(lambda x: (x[x==0]).count()/(x.count()))
	return fractions


def count_records(trajectories_frame):
	"""
	Returns total number of records for each user

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		A pandas Series with a count of records for each user
	"""
	counts = trajectories_frame.groupby(level=0).agg('count').iloc[:, 0]
	return counts


def count_records_per_time_frame(trajectories_frame, resolution):
	"""
	Returns total number of records for each user per time frame

	Args:
		trajectories_frame: TrajectoriesFrame class object
		resolution: a time frame per which the count will be calculated

	Returns:
		A pandas Series with a count of records for each user
	"""
	level_values = trajectories_frame.index.get_level_values
	time_frame_counts = trajectories_frame.groupby([level_values(0)] + [pd.Grouper(freq=resolution, level=-1)]).agg('count').iloc[:, 0]
	return time_frame_counts


def user_trajectories_duration(trajectories_frame, resolution, count_empty=True):
	"""
	Returns the total duration of users' trajectories

	Args:
		trajectories_frame: TrajectoriesFrame class object
		resolution: time unit in which duration will be expressed
		count_empty: if empty records should be included

	Returns:
		a pandas Series with the duration of each user's trajectory
	"""
	total_time_duration = trajectories_frame.groupby(level=0).resample(resolution, level=1).count().iloc[:, 0]
	if count_empty:
		total_time_duration = total_time_duration.groupby(level=0).count()
	else:
		total_time_duration = total_time_duration[total_time_duration >= 1].groupby(level=0).count()
	return total_time_duration


def consecutive_record(trajectories_frame, resolution):
	"""
	Calculates the maximum length of consecutive records for each user

	Args:
		trajectories_frame: TrajectoriesFrame object class
		resolution: time unit at which the consecutive records will be counted

	Returns:
		a pandas Series with the maximum length of consecutive records for each user
	"""
	total_time_duration = trajectories_frame.groupby(level=0).resample(resolution, level=1).count().iloc[:, 0]
	total_time_duration[total_time_duration > 0] = 1
	total_time_duration = total_time_duration.groupby(level=0).progress_apply(
		lambda x: x * x.groupby((total_time_duration != total_time_duration.shift()).cumsum()).cumcount() + 1)
	total_time_duration = total_time_duration.groupby(level=0).max()
	return total_time_duration


def get_max(statistical_series):
	"""
	Get maximum value for each group

	Args:
		statistical_series: Multiindex series

	Returns:
		A series with maximum value for each group
	"""
	return statistical_series.groupby(level=0).agg('max')


def get_min(statistical_series):
	"""
	Get minimum value for each group

	Args:
		statistical_series: Multiindex series

	Returns:
		A series with minimum value for each group
	"""
	return statistical_series.groupby(level=0).agg('min')


def get_mean(statistical_series):
	"""
	Get mean value for each group

	Args:
		statistical_series: Multiindex series

	Returns:
		A series with mean value for each group
	"""
	return statistical_series.groupby(level=0).agg('mean')



import numpy as np


def where(user):
    """
    WHERE module of the model
    :param user: currently processed user
    """
    try:
        user.history.append(user.sig_locs[user.current_loc])
    except KeyError:
        try:
            user.history.append(np.random.choice([user.activity_area[n][0] for n in range(len(user.activity_area))]).centroid)
        except KeyError:
            user.history.append(user.sig_locs[0])


from humobi.misc.utils import resolution_to_points, moving_average
import numpy as np
from math import ceil
from shapely import geometry
import geopandas as gpd


def create_grid(trajectories_frame, resolution):
	"""
	Creates grid fitting given trajectories in TrajectoriesFrame class object.

	Args:
		trajectories_frame: TrajectoriesFrame class object
		resolution: grid resolution

	Returns:
		A GeoDataFrame with the grid
	"""
	crs= trajectories_frame.crs
	_x_min = trajectories_frame['geometry'].x.min()
	_x_max = trajectories_frame['geometry'].x.max()
	_y_min = trajectories_frame['geometry'].y.min()
	_y_max = trajectories_frame['geometry'].y.max()
	x_points = resolution_to_points(resolution, _x_max, _x_min)
	y_points = resolution_to_points(resolution, _y_max, _y_min)
	x = np.linspace(_x_min, _x_max, ceil(x_points))
	y = np.linspace(_y_min, _y_max, ceil(y_points))
	grid_cells = []
	for xn in range(len(x)-1):
		for yn in range(len(y)-1):
			grid_cells.append(geometry.box(x[xn], y[yn], x[xn+1], y[yn+1]))
	return gpd.GeoDataFrame(grid_cells, columns=['geometry'], crs=crs)



# IMPORTS
from humobi.preprocessing.temporal_aggregation import TemporalAggregator
import numpy as np
import pandas as pd
from humobi.tools.processing import groupwise_normalise, groupwise_expansion
from humobi.misc.utils import *
from humobi.misc.utils import _repeatfinder_dense, _repeatfinder_sparse, _repeatfinder_equally_sparse, \
	_global_align, _iterative_global_align
from tqdm import tqdm
from humobi.structures.trajectory import TrajectoriesFrame
tqdm.pandas()
import concurrent.futures as cf
from math import ceil
from random import sample
from scipy.optimize import curve_fit
from sklearn.metrics import r2_score


def num_of_distinct_locations(trajectories_frame):
	"""
	Returns a number of distinct location in the trajectory. First looks for 'labels' column.

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a Series with the number of unique locations for each user
	"""
	if isinstance(trajectories_frame, pd.DataFrame):
		return trajectories_frame.groupby(level=0).progress_apply(lambda x: len(pd.unique(x['labels'])))
	else:
		return trajectories_frame.groupby(level=0).progress_apply(lambda x: pd.unique(x).shape[0])


def visitation_frequency(trajectories_frame):
	"""
	Calculates visitiation frequency for each user in the TrajectoriesFrame.

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a Series with the visitation frequency for each user and for each location. Sorted descendingly.
	"""
	lat_col = trajectories_frame._geom_cols[0]
	lon_col = trajectories_frame._geom_cols[1]
	frequencies = trajectories_frame.groupby(level=0).progress_apply(
		lambda x: x.groupby([lat_col, lon_col]).count()).iloc[:, 0]
	frequencies = frequencies.groupby(level=0).progress_apply(lambda x: x.sort_values(ascending=False))
	frequencies = groupwise_normalise(frequencies)
	return frequencies


def _filter_distinct_locations(trajectories_frame):
	to_concat = []
	for ind, vals in tqdm(trajectories_frame.groupby(level=0), total=len(trajectories_frame.groupby(level=0)) ):
		if len(vals) == 1:
			to_concat.append(uniq)
			continue
		else:
			uniq = (vals['geometry'][vals['geometry'].is_valid].drop_duplicates())
			to_concat.append(uniq)
	return pd.concat(to_concat)


def distinct_locations_over_time(trajectories_frame, resolution='1H', reaggregate=False):
	"""
	Calculates the number of distinct location visited in the movement trajectory over time.

	Args:
		trajectories_frame: TrajectoriesFrame class object
		resolution: determines the time unit for the metric
		reaggregate: if true, data are first reagregated to given time unit

	Returns:
		a Series with the number of unique locations visited up to each time step in the movement trajectory
	"""
	if reaggregate:
		temp_agg = TemporalAggregator(resolution)
		trajectories_frame = temp_agg.aggregate(trajectories_frame)
	trajectories_frame = _filter_distinct_locations(trajectories_frame)
	distinct_locations = trajectories_frame.groupby(level=0).resample(resolution, level=1).count()
	distinct_locations = distinct_locations.groupby(level=0).cumsum()
	return distinct_locations


def jump_lengths(trajectories_frame):
	"""
	Calculates jump lengths between each step in the trajectory

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a Series with jump lengths between consecutive records
	"""
	jumps = trajectories_frame.dropna().geometry.groupby(level=0).progress_apply(lambda x: x.distance(x.shift()))
	return jumps


def nonzero_trips(trajectories_frame):
	"""
	Counts all trips that had distance larger than 0.

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a Series with a count of nonzero trips for each user
	"""
	jumps = jump_lengths(trajectories_frame).dropna().droplevel(1)
	return jumps[jumps != 0].groupby(level=0).count()


def self_transitions(trajectories_frame):
	"""
	Calculates the number of self transitions for each user

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Retruns
		a Series with the number of self transitions for each user
	"""
	if isinstance(trajectories_frame, pd.Series):
		self_transitions_mask = (trajectories_frame == trajectories_frame.shift())
	else:
		if not hasattr(trajectories_frame, '_geom_cols'):
			trajectories_frame = TrajectoriesFrame(trajectories_frame)
		coordinates_frame = trajectories_frame[[trajectories_frame._geom_cols[0], trajectories_frame._geom_cols[1]]]
		self_transitions_mask = (coordinates_frame == coordinates_frame.shift()).all(axis=1)
	empty_mask = (~self_transitions_mask).groupby(level=0).progress_apply(lambda x: x.all())
	empty_mask = empty_mask[empty_mask == True].index
	self_transitions_only = trajectories_frame[self_transitions_mask]
	empty_self_transitions = pd.DataFrame([0 for x in range(len(empty_mask))], index=empty_mask)
	if isinstance(trajectories_frame, pd.Series):
		self_transitions_only = self_transitions_only.groupby(level=0).count()
	else:
		self_transitions_only = self_transitions_only.groupby(level=0).count()[self_transitions_only.columns[0]]
	if len(empty_self_transitions) > 0:
		self_transitions_only.append(empty_self_transitions.iloc[:, 0]).sort_index()
	return self_transitions_only


def waiting_times(trajectories_frame, time_unit='h'):
	"""
	Calculates waiting times for each transition in TrajectoriesFrame

	Args:
		trajectories_frame: TrajectoriesFrame class object
		time_unit: time unit in which waiting times will be expressed

	Returns:
		A series with waiting times for each transition for each user
	"""
	trajectories_frame = trajectories_frame.dropna()
	transitions_only = trajectories_frame[
		trajectories_frame.geometry.groupby(level=0).progress_apply(lambda x: x.shift(-1) != x)]
	transitions_only['dt'] = transitions_only.index.get_level_values(1)
	times = transitions_only.groupby(level=0).dt.progress_apply(lambda x: (x - x.shift()).astype('timedelta64[%s]' % time_unit))
	return times


def center_of_mass(trajectories_frame):
	"""
	Calculates a center of mass for each user's trajectory

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a GeoSeries with centers of mass of each user's trajectory
	"""
	return trajectories_frame.dissolve(by=trajectories_frame.index.get_level_values(0)).centroid


def radius_of_gyration(trajectories_frame, time_evolution=True):
	"""
	Calculates radii of gyration for each user. Optionally uses time steps to express their growth.

	Args:
		trajectories_frame: TrajectoriesFrame class object
		time_evolution: If true, radii of gyration are calculated over time

	Returns:
		a Series with radii of gyration for each user
	"""
	mean_locs = center_of_mass(trajectories_frame)
	to_concat_dict = {}
	to_concat_list = []
	for ind, vals in tqdm(trajectories_frame.groupby(level=0), total=len(trajectories_frame.groupby(level=0))):
		vals = vals.dropna()
		rog_ind = vals.distance(mean_locs.loc[ind]) ** 2
		if time_evolution:
			rog_ind = groupwise_expansion(np.sqrt(rog_ind))
			to_concat_list.append(rog_ind)
		else:
			rog_ind = np.sqrt(rog_ind.mean())
			to_concat_dict[ind] = rog_ind
	if time_evolution:
		radius = pd.concat(to_concat_list)
	else:
		radius = pd.DataFrame.from_dict(to_concat_dict, orient='index')
	return radius


def mean_square_displacement(trajectories_frame, from_center=False, time_evolution=True, reference_locs=None):
	"""
	Calculates mean square displacements for each user. Optionally uses time steps to express their growth.

	Args:
		trajectories_frame: TrajectoriesFrame class object
		from_center: If true, displacement is calculated from the trajectory ceneter, if false - from the first point
		time_evolution: If true, mean square displacements are calculated over time
		reference_locs: allows to give reference locations for each trajectory explicitly

	Returns:
		a Series with mean square displacements for each user
	"""
	to_concat_dict = {}
	to_concat_list = []
	if reference_locs is None:
		if from_center:
			reference_locs = center_of_mass(trajectories_frame)
		else:
			reference_locs = trajectories_frame.groupby(level=0).head(1).droplevel(1).geometry
	for ind, vals in tqdm(trajectories_frame.groupby(level=0), total=len(trajectories_frame.groupby(level=0))):
		vals = vals.dropna()
		msd_ind = (vals.distance(reference_locs.loc[ind]) ** 2)
		if time_evolution:
			msd_ind = groupwise_expansion(msd_ind)
			to_concat_list.append(msd_ind)
		else:
			msd_ind = msd_ind.mean()
			to_concat_dict[ind] = msd_ind
	if time_evolution:
		msd = pd.concat(to_concat_list)
	else:
		msd = pd.DataFrame.from_dict(to_concat_dict, orient='index')
	return msd


def return_time(trajectories_frame, time_unit='h', by_place=False):
	"""
	Calculates return times for each unique location in each user's trajectory.

	Args:
		trajectories_frame: TrajectoriesFrame class object
		time_unit: time unit in which return times will be expressed
		by_place: If true, return times are expressed for each place globally

	Returns:
		a Series with return times
	"""
	if not hasattr(trajectories_frame, '_geom_cols'):
		trajectories_frame = TrajectoriesFrame(trajectories_frame)
	lat_col = trajectories_frame[trajectories_frame._geom_cols[0]]
	lon_col = trajectories_frame[trajectories_frame._geom_cols[1]]
	trajectories_frame['datetime_temp'] = trajectories_frame.index.get_level_values(1)
	to_concat = []
	for ind, vals in tqdm(trajectories_frame.groupby(level=0), total=len(trajectories_frame.groupby(level=0))):
		concat_level = {}
		for place, vals2 in vals.groupby([lat_col, lon_col]):
			shifts = (vals2.datetime_temp - vals2.datetime_temp.shift()).astype('timedelta64[%s]' % time_unit)
			concat_level[place] = shifts
		to_concat.append(pd.concat(concat_level))
	return_times = pd.concat(to_concat)
	if by_place:
		return_times = return_times.groupby(level=2).progress_apply(
			lambda x: x.groupby(level=[0, 1]).agg(['count', 'mean']).dropna())
		return_times = return_times.groupby(level=0).progress_apply(lambda x: x.sort_values('count', ascending=False))
	else:
		return_times = return_times.groupby(level=2).progress_apply(lambda x: x.sort_values(ascending=False)).droplevel(
			[1, 2])
	return return_times


def random_entropy(trajectories_frame):
	"""
	Calculates random entropy for each user in TrajectoriesFrame

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a Series with random entropies for each user
	"""
	return trajectories_frame.groupby(level=0).progress_apply(lambda x: np.log2(len(pd.unique(x.geometry))))


def unc_entropy(trajectories_frame):
	"""
	Calculates uncorrelated entropy for each user in TrajectoriesFrame

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a Series with uncorrelated entropies for each user
	"""
	frequencies = visitation_frequency(trajectories_frame)
	return frequencies.groupby(level=0).progress_apply(lambda x: -np.sum([pk * np.log2(pk) for pk in x if pk != 0]))


def _fit_func(x, a, b, c):
	return a * np.exp(b * x) + c


def _real_entropy(indi, gs):
	"""
	Calculates actual entropy for series of symbols

	Args:
		indi: unique identifier
		gs: series of symbols

	Returns:
		a unique identifier and entropy value
	"""
	return indi, np.power(np.mean(matchfinder(gs)), -1) * np.log2(len(gs))


def _real_scalling_entropy(indi, trace, estimation_method='unc'):
	"""
	Calculates actual entropy for trajectories. If trajectory has missing data, uses estimation. Uncorrelated entropy-based
	estimation is used.

	Args:
		indi: unique identifier
		trace: movement trajectory

	Returns:
		a unique identifier and actual entropy
	"""
	empty_fraction = trace.isnull().sum() / trace.shape[0]
	if empty_fraction < .15:
		return _real_entropy(indi, trace)
	elif empty_fraction >.9:
		raise ValueError("Fraction of missing records (q = %f) to high to estimate real entropy!" % empty_fraction)
	estimation_step = ceil((.9 - empty_fraction) / .05)
	range_to_empty = [empty_fraction + .05 * x for x in range(estimation_step)]
	scaling_features = []
	uncs = []
	real_qs = []
	if estimation_method == 'unc':
		visit_freq = trace.value_counts() / trace.shape[0]  # FOR UNCORRELATED ENTROPY-BASED ESTIMATION
		Sunc_baseline = -np.sum(visit_freq * np.log2(visit_freq))  # FOR UNCORRELATED ENTROPY-BASED ESTIMATION
	elif estimation_method == 'shuff':
		Sunc_baseline = np.power(np.mean(matchfinder(trace.sample(frac=1).reset_index(drop=True))), -1) * \
		                np.log2(len(
			                trace.sample(frac=1).reset_index(drop=True)))  # FOR SHUFFLED TRAJECTORY-BASED ESTIMATION
	for q in range_to_empty[1:]:
		trace_copy2 = trace.copy()
		points_to_remove = sample(set(trace_copy2[~trace_copy2.isnull()].index),
		                          int(round((q - empty_fraction) * len(trace_copy2))))
		trace_copy2.loc[points_to_remove] = None
		Strue = np.power(np.mean(matchfinder(trace_copy2)), -1) * np.log2(len(trace_copy2))
		trace_shuffled = trace_copy2.sample(frac=1).reset_index(drop=True)
		if estimation_method == 'unc':
			visit_freq = trace_copy2.value_counts() / trace_copy2.shape[0]  # FOR UNCORRELATED ENTROPY-BASED ESTIMATION
			Sunc = -np.sum(visit_freq * np.log2(visit_freq))  # FOR UNCORRELATED ENTROPY-BASED ESTIMATION
		elif estimation_method == 'shuff':
			Sunc = np.power(np.mean(matchfinder(trace_shuffled)), -1) * np.log2(len(trace_shuffled)) # FOR SHUFFLED TRAJECTORY-BASED ESTIMATION
		scaling_features.append(np.log2(Strue / Sunc))
		uncs.append(Sunc)
		real_qs.append(sum(trace_copy2.isnull()) / len(trace_copy2))
	try:
		popt, pcov = curve_fit(_fit_func, real_qs, scaling_features, maxfev=12000, p0=[0.1, 2, 0.1])
		if sum(scaling_features) == 0 and r2_score(scaling_features, [_fit_func(x, *popt) for x in real_qs]) < .9:
			a, b = np.polyfit(real_qs, scaling_features, 1)
			return indi, np.power(2, b) * Sunc_baseline
		else:
			return indi, np.power(2, _fit_func(0, *popt)) * Sunc_baseline
	except:
		a, b = np.polyfit(real_qs, scaling_features, 1)
		return indi, np.power(2, b) * Sunc_baseline


def real_entropy(trajectories_frame):
	"""
	Calculates an actual entropy for each user in TrajectoriesFrame

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a Series with actual entropies for each user
	"""
	result_dic = {}
	with cf.ThreadPoolExecutor() as executor:
		try:
			args = [val.labels for indi, val in trajectories_frame.groupby(level=0)]
		except KeyError:
			args = [val for indi, val in trajectories_frame.groupby(level=0)]
		ids = [indi for indi, val in trajectories_frame.groupby(level=0)]
		results = list(tqdm(executor.map(_real_scalling_entropy, ids, args), total=len(ids)))
	for result in results:
		result_dic[result[0]] = result[1]
	return pd.Series(np.fromiter(result_dic.values(), dtype=float), index=np.fromiter(result_dic.keys(), dtype=int))


def random_predictability(trajectories_frame):
	"""
	Calculates random entropy and predictability.

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a Series with random entropy and predictability for each user
	"""
	distinct_locations = num_of_distinct_locations(trajectories_frame)
	rand_ent = random_entropy(trajectories_frame)
	merged = pd.DataFrame([distinct_locations, rand_ent], index=['locations', 'entropy'])
	return merged.progress_apply(lambda x: fano_inequality(x['locations'], x['entropy'])), rand_ent


def unc_predictability(trajectories_frame):
	"""
	Calculates uncorrelated entropy and predictability.

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a Series with uncorrelated entropy and predictability for each user
	"""
	distinct_locations = num_of_distinct_locations(trajectories_frame)
	unc_ent = unc_entropy(trajectories_frame)
	merged = pd.DataFrame([distinct_locations, unc_ent], index=['locations', 'entropy'])
	return merged.progress_apply(lambda x: fano_inequality(x['locations'], x['entropy'])), unc_ent


def real_predictability(trajectories_frame):
	"""
	Calculates actual entropy and predictability.

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a Series with actual entropy and predictability for each user
	"""
	distinct_locations = num_of_distinct_locations(trajectories_frame)
	real_ent = real_entropy(trajectories_frame)
	merged = pd.DataFrame([distinct_locations, real_ent], index=['locations', 'entropy'])
	return merged.progress_apply(lambda x: fano_inequality(x['locations'], x['entropy'])), real_ent


def stationarity(trajectories_frame):
	"""
	Calculates the stationarity according to Teixeira et al. (2019) as the average stay length in the location.

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a Series with stationarity values for each user
	"""
	trajectories_frame = trajectories_frame.dropna()
	stationarity_frame = trajectories_frame.labels.groupby(level=0).apply(
		lambda x: x.groupby((x != x.shift()).cumsum()).size()-1)
	stationarity_frame = stationarity_frame.groupby(level=0).sum()
	size_frame = trajectories_frame.groupby(level=0).apply(lambda x: x.size)
	return stationarity_frame/size_frame


def regularity(trajectories_frame):
	"""
	Calculates the regularity according to Teixeira et al. (2019) as the ratio of sequence lenght and the number of
	unique symbols

	Args:
		trajectories_frame: TrajectoriesFrame class object

	Returns:
		a Series with regularity values for each user
	"""
	regul = trajectories_frame.groupby(level=0).progress_apply(lambda x: len(x)) / num_of_distinct_locations(
		trajectories_frame)
	return regul


def _repeatability_dense(indi, S1, S2):
	S1 = S1.values
	S2 = S2.values
	return indi, _repeatfinder_dense(S1, S2)


def _repeatability_sparse(indi, S1, S2):
	S1 = S1.values
	S2 = S2.values
	return indi, _repeatfinder_sparse(S1, S2)


def _repeatability_equally_sparse(indi, S1, S2):
	S1 = S1.values
	S2 = S2.values
	return indi, _repeatfinder_equally_sparse(S1,S2)


def _global_alignment(indi, S1, S2):
	S1 = S1.values
	S2 = S2.values
	return indi, _global_align(S1, S2)


def _iterative_global_alignment(indi, S1, S2):
	S1 = S1.values
	S2 = S2.values
	return indi, _iterative_global_align(S1, S2)


def repeatability_sparse(train_frame, test_frame):
	"""
	Calculates the sparse repeatability metric proposed in the paper Smolak et al. (2022) as the number of
	transitions repeating in two sequences in the same order. Matched transitions can be separated by a gap of any size.
	Returns normalised values.

	Args:
		train_frame: TrajectoriesFrame class object with the training data
		test_frame:TrajectoriesFrame class object with the test data

	Returns:
		a Series with sparse repeatability values for each user
	"""
	result_dic = {}
	with cf.ThreadPoolExecutor() as executor:
		args_train = [val.labels for indi, val in train_frame.groupby(level=0)]
		args_test = [val.labels for indi, val in test_frame.groupby(level=0)]
		ids = [indi for indi, val in test_frame.groupby(level=0)]
		results = list(tqdm(executor.map(_repeatability_sparse, ids, args_train, args_test), total=len(ids)))
	for result in results:
		result_dic[result[0]] = result[1]
	return pd.Series(np.fromiter(result_dic.values(), dtype=float), index = np.fromiter(result_dic.keys(), dtype = int))


def repeatability_dense(train_frame, test_frame):
	"""
	Calculates the dense repeatability metric proposed in the paper Smolak et al. (2022) as the number of
	transitions repeating in two sequences in the same order. Matched transitions cannot be separated by a gap.
	Returns normalised values.

	Args:
		train_frame: TrajectoriesFrame class object with the training data
		test_frame:TrajectoriesFrame class object with the test data

	Returns:
		a Series with dense repeatability values for each user
	"""
	result_dic = {}
	with cf.ThreadPoolExecutor() as executor:
		args_train = [val.labels for indi, val in train_frame.groupby(level=0)]
		args_test = [val.labels for indi, val in test_frame.groupby(level=0)]
		ids = [indi for indi, val in test_frame.groupby(level=0)]
		results = list(tqdm(executor.map(_repeatability_dense, ids, args_train, args_test), total=len(ids)))
	for result in results:
		result_dic[result[0]] = result[1]
	return pd.Series(np.fromiter(result_dic.values(), dtype=float), index=np.fromiter(result_dic.keys(), dtype=int))


def repeatability_equally_sparse(train_frame, test_frame):
	"""
	Calculates the equally sparse repeatability metric proposed in the paper Smolak et al. (2022) as the number of
	transitions repeating in two sequences in the same order. Matched transitions can be separated by a gap of any size.
	Matched transitions HAVE TO BE spearated by the gap of the same size.
	Returns normalised values.

	Args:
		train_frame: TrajectoriesFrame class object with the training data
		test_frame:TrajectoriesFrame class object with the test data

	Returns:
		a Series with equally sparse repeatability values for each user
	"""
	result_dic = {}
	with cf.ThreadPoolExecutor() as executor:
		args_train = [val.labels for indi, val in train_frame.groupby(level=0)]
		args_test = [val.labels for indi, val in test_frame.groupby(level=0)]
		ids = [indi for indi, val in test_frame.groupby(level=0)]
		results = list(tqdm(executor.map(_repeatability_equally_sparse, ids, args_train, args_test), total=len(ids)))
	for result in results:
		result_dic[result[0]] = result[1]
	return pd.Series(np.fromiter(result_dic.values(), dtype=float), index = np.fromiter(result_dic.keys(), dtype = int))


def global_alignment(train_frame, test_frame):
	"""
	Calculates the global alignment metric proposed in the paper Smolak et al. (2022) through the application of the
	Needleman-Wunsch algorithm on sequences for pairwise matching.
	Returns normalised values.

	Args:
		train_frame: TrajectoriesFrame class object with the training data
		test_frame: TrajectoriesFrame class object with the test data

	Returns:
		a Series with global alginment metric values for each user
	"""
	result_dic = {}
	with cf.ThreadPoolExecutor() as executor:
		args_train = [val.labels for indi, val in train_frame.groupby(level=0)]
		args_test = [val.labels for indi, val in test_frame.groupby(level=0)]
		ids = [indi for indi, val in test_frame.groupby(level=0)]
		results = list(tqdm(executor.map(_global_alignment, ids, args_train, args_test), total=len(ids)))
	for result in results:
		result_dic[result[0]] = result[1]
	return pd.Series(np.fromiter(result_dic.values(), dtype=float), index = np.fromiter(result_dic.keys(), dtype = int))


def iterative_global_alignment(train_frame, test_frame):
	"""
	Calculates the iterative global alignment metric proposed in the paper Smolak et al. (2022) through the application 
	of the Needleman-Wunsch algorithm on sequences for pairwise matching with the iterative approach.
	Returns normalised values.

	Args:
		train_frame: TrajectoriesFrame class object with the training data
		test_frame: TrajectoriesFrame class object with the test data

	Returns:
		a Series with global alignment metric values for each user
	"""
	result_dic = {}
	with cf.ThreadPoolExecutor() as executor:
		args_train = [val.labels for indi, val in train_frame.groupby(level=0)]
		args_test = [val.labels for indi, val in test_frame.groupby(level=0)]
		ids = [indi for indi, val in test_frame.groupby(level=0)]
		results = list(tqdm(executor.map(_iterative_global_alignment, ids, args_train, args_test), total=len(ids)))
	for result in results:
		result_dic[result[0]] = result[1]
	return pd.Series(np.fromiter(result_dic.values(), dtype=float), index = np.fromiter(result_dic.keys(), dtype = int))



import numpy as np
from humobi.models.spatial_tools.misc import normalize_array
WEIGHT = False


def when(user,slot,circadian_rhythm):
	"""
	WHEN module of the model
	:param user: Currently processed user
	:param slot: Currently processed time slot
	:param circadian_rhythm: Circadian rhythm of the group within which the user is
	"""
	if WEIGHT:
		probabilities = normalize_array(np.array([circadian_rhythm[user.cluster][slot.hour],
												circadian_rhythm[user.cluster][slot.hour + 24],
												circadian_rhythm[user.cluster][slot.hour + 48]]))
		user.current_loc = np.random.choice([0, 1, 2], size=1, p=probabilities)[0]
	if not WEIGHT:
		user.current_loc = int(circadian_rhythm[user.cluster][slot.hour])


import tensorflow as tf
import numpy as np
import pandas as pd
from sklearn.model_selection import KFold


def gen(x_data, n_splits):
	"""
	Generates data for learning.

	Args:
		x_data: DataFrame of data for learning
		n_splits: the number of splits

	Returns:
		training and testing datasets
	"""
	for train_index, test_index in KFold(n_splits).split(x_data):
		x_train, x_test = x_data[train_index], x_data[test_index]
		yield x_train, x_test


def split_input_target(chunk):
	"""
	Splits data into chunks of inputs and target values.

	Args:
		chunk: chunk to split

	Returns:
		split data
	"""
	# for the example: hello
	input_text = chunk[:-1]  # hell
	target_text = chunk[1:]  # ello
	return input_text, target_text  # hell, ello


class GRUModel2(tf.keras.Model):
	"""
	GRU Model (2 GRU layers + embedding + dropout layer)

	Args:
		vocab_size: Vocabulary size - the number of unique places in sequence
		embedding_dim: The size of embedding layer
		rnn_units: The number of rnn units on a hidden layer
		dropout: Dropout size
		batch_size: Size of a single batch
	"""

	def __init__(self, vocab_size, embedding_dim, rnn_units, dropout, batch_size, window_size):
		"""
		Layer structure.
		"""
		super().__init__(self)
		self.embedding = tf.keras.layers.Embedding(vocab_size, embedding_dim, batch_input_shape=(batch_size,vocab_size))
		self.gru = tf.keras.layers.GRU(rnn_units,
		                               return_sequences=True,
		                               return_state=True, batch_input_shape=(batch_size, window_size, embedding_dim))
		self.gru2 = tf.keras.layers.GRU(rnn_units,
		                                return_sequences=True,
		                                return_state=True, batch_input_shape=(batch_size, window_size, embedding_dim))
		self.drop = tf.keras.layers.Dropout(dropout)
		self.dense = tf.keras.layers.Dense(vocab_size, activation='softmax')

	def call(self, inputs, states=None, return_state=False, training=False):
		"""
		Constructs the network for training.

		Args:
			inputs: Input vectors
			states: Initial states. If none, no initial state is set.
			return_state (default = False): Boolean. Whether to return the last state in addition to the output.
			training (default = False): Boolean. Indicating whether the layer should behave in training mode or in inference mode

		Returns:
			The GRU network
		"""
		x = inputs
		x = self.embedding(x)
		if states is None:
			states = self.gru.get_initial_state(x)
		x, states = self.gru(x, initial_state=states, training=training)
		x, states = self.gru2(x, initial_state=states, training=training)
		x = self.drop(x, training=training)
		x = self.dense(x)

		if return_state:
			return x, states
		else:
			return x


class GRUModel(tf.keras.Model):
	"""
		GRU Model (1 GRU layer + embedding + dropout layer)

	Args:
		vocab_size: Vocabulary size - the number of unique places in sequence
		embedding_dim: The size of embedding layer
		rnn_units: The number of rnn units on a hidden layer
		dropout: Dropout size
		batch_size: Size of a single batch
	"""

	def __init__(self, vocab_size, embedding_dim, rnn_units, dropout, batch_size):
		super().__init__(self)
		self.embedding = tf.keras.layers.Embedding(vocab_size, embedding_dim)
		self.gru = tf.keras.layers.GRU(rnn_units,
		                               return_sequences=True,
		                               return_state=True, stateful=True)
		self.drop = tf.keras.layers.Dropout(dropout)
		self.dense = tf.keras.layers.Dense(vocab_size, activation='softmax')

	def call(self, inputs, states=None, return_state=False, training=False):
		"""
		Constructs the network for training.

		Args:
			inputs: Input vectors
			states: Initial states. If none, no initial state is set.
			return_state (default = False): Boolean. Whether to return the last state in addition to the output.
			training (default = False): Boolean. Indicating whether the layer should behave in training mode or in inference mode

		Returns:
			The GRU network
		"""
		x = inputs
		x = self.embedding(x)
		if states is None:
			states = self.gru.get_initial_state(x)
		x, states = self.gru(x, initial_state=states, training=training)
		x = self.drop(x, training=training)
		x = self.dense(x, training=training)

		if return_state:
			return x, states
		else:
			return x


def loss(labels, logits):
	"""
	The loss function - categorical crossentropy.

	Args:
		labels: ground truth values
		logits: predictions encoded as probability distribution

	Returns:
		loss values
	"""
	return tf.keras.losses.sparse_categorical_crossentropy(labels, logits, from_logits=True)


class DeepPred():
	"""
	Deep-learning based sequence prediction model

	Args:
		model: Name of the model to use. Available: GRU (single layer), GRU2 (GRU double layer), LSTM (single layer)
		data: Data for prediction - whole dataset, including test. Do not split it.
		test_size: Test ratio for data split
		folds: The number of folds in cross-validation of the model
		window_size: Window size, also referenced as the size of a lookback. The number of previous symbols considered
		during prediction.
		batch_size: The size of a batch for a single network
		embedding_dim: The size of an embedding layer (the number of nodes)
		rnn_units:  The number of RNN layer's units
	"""

	def __init__(self, model, data, test_size=.2, folds=5, window_size=2, batch_size=1, embedding_dim=1024,
	             rnn_units=1024):
		"""
		Class initialisation. Runs data preparation - slicing into training and test sets.
		"""
		self.model = model
		self.data = data
		self.split_ratio = test_size
		self.folds = folds
		self.window_size = window_size
		self.batch_size = batch_size
		self.embedding_dim = embedding_dim
		self.rnn_units = rnn_units
		self.scores = {}
		self.predictions = {}
		self._prepare()

	def _prepare(self):
		"""
		Prepares data for prediction - splits data into train and test sets.
		"""
		data_len = self.data.groupby(level=0).apply(lambda x: len(x))
		cut_index = (1-round(self.split_ratio * data_len)).astype(int)
		data_train = self.data.groupby(level=0).apply(
			lambda x: x.labels.iloc[:cut_index.loc[x.index.get_level_values(0)[0]]])
		data_test = self.data.groupby(level=0).apply(
			lambda x: x.labels.iloc[cut_index.loc[x.index.get_level_values(0)[0]] - self.window_size:])
		self.data = (data_train, data_test)

	def _user_learn(self, tr_data, ts_data):
		ts_dataset = tf.data.Dataset.from_tensor_slices(ts_data)  # transform test set into tensor
		if self.folds != 1:  # if there is more than single fold
			train_folds = [(tf.data.Dataset.from_tensor_slices(x), tf.data.Dataset.from_tensor_slices(y)) for x, y in
			               gen(tr_data, self.folds)]  # then split data into the list folds
		else:  # if there is a single fold
			train_folds = [[(tf.data.Dataset.from_tensor_slices(x), tf.data.Dataset.from_tensor_slices(y)) for x, y in
			                gen(tr_data, 2)][0]]  # make it a one set of data
		self.s_regions = np.hstack([tr_data, ts_data]).max() + 1  # read the number of unique symbols from the dataset
		sequences_train = [dataset_train[0].batch(self.window_size + 1, drop_remainder=True) for dataset_train in
		                   train_folds]  # prepare the batches of sequences for training
		sequences_val = [dataset_train[1].batch(self.window_size + 1, drop_remainder=True) for dataset_train in
		                 train_folds]  # prepare the batches of sequences for validation
		sequences_test = ts_dataset.batch(self.window_size + 1, drop_remainder=True)  # prepare the batches of
		# sequences for tests
		dataset_train = [t.map(split_input_target) for t in sequences_train]  # and convert them using the windowing algorithm
		dataset_val = [v.map(split_input_target) for v in sequences_val]
		dataset_test = sequences_test.map(split_input_target)  # data test is split but not used - just for debugging
		data_tr = [t.shuffle(1000).batch(self.batch_size, drop_remainder=True) for t in
		           dataset_train]  # shuffle the data for training and batch
		data_val = [v.shuffle(1000).batch(self.batch_size, drop_remainder=True) for v in
		            dataset_val]  # shuffle the data for testing and batch
		fold = 1  # the counter of folds
		callback = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=2)  # initialise the EarlyStopping mechanism
		# Initialiase selected models - the dropout rate is fixed
		if self.model == "GRU":
			model = GRUModel(self.s_regions, self.embedding_dim, self.rnn_units, 0.2, self.batch_size)
		elif self.model == "GRU2":
			model = GRUModel2(self.s_regions, self.embedding_dim, self.rnn_units, 0.2, self.batch_size, self.window_size)
		model.compile(optimizer='adam', loss=loss, metrics=['accuracy'])  # compile the model
		for x, v in zip(data_tr, data_val):  # the CV training process
			print("FOLD:{}".format(fold))
			self.history = model.fit(x, epochs=30, validation_data=v, callbacks=[callback], batch_size=self.batch_size)
			fold += 1
		model.reset_states()
		return model

	def learn_predict(self):
		"""
		Trains the network and makes prediction on the test set. The prediction is given a fixed temperature = 0.1, you
		can change it here. The higher the more random are predictions. However, small temperature can cause network
		to stuck in an infinite loop.
		"""
		result_dic = {}
		predictions_dic = {}
		train = self.data[0].groupby(level=0)
		test = self.data[1].groupby(level=0)
		for tr, ts in zip(train, test):
			uid = tr[0]
			tr_data = tr[1].values.astype(int)
			ts_data = ts[1].values.astype(int)
			user_model = self._user_learn(tr_data, ts_data)
			stabs = []
			temperature = .1 # TEMPERATURE
			for stab in range(1): # TODO: add as parameter
				forecast = []
				for x in range(len(ts_data) - self.window_size):
					y = tf.expand_dims(ts_data[x:x + self.window_size], 0)
					predictions = user_model.predict(y, batch_size=self.batch_size)
					predictions = np.squeeze(predictions, 0)
					predictions = predictions / temperature
					predicted_id = tf.random.categorical(predictions, num_samples=1)[-1, 0].numpy()
					forecast.append(predicted_id)
				stabs.append(sum(forecast == ts_data[self.window_size:]) / len(forecast))
			predictions_dic[uid] = (forecast,ts_data[self.window_size:])
			print(np.mean(stabs))
			result_dic[uid] = np.mean(stabs)
		self.scores = pd.DataFrame.from_dict(result_dic, orient='index')
		self.predictions = pd.concat({k: pd.concat([pd.Series(v[0]),pd.Series(v[1])],axis=1) for k,v in predictions_dic.items()}).droplevel(1)



from humobi.models.agent_module.agent_class import Agent


def _generate_agent(cluster_id, home, work, area):
	sig_locations = (home, work)
	return Agent(sig_locs=sig_locations, temporal_mechanism=None, spatial_mechanism=None,
	             cluster=cluster_id, activity_area=area)


def generate_agents(n_agents, cluster_id, list_home, list_work, list_geometry):
	"""
	Returns a list of generated Agent class objects.

	Args:
		n_agents: the number of agents to generate
		cluster_id: id of a cluster
		list_home: contains home positions
		list_work: contains work positions
		list_geometry: contains activity areas

	Returns:
		a list of generated agents
	"""
	agents = []
	for n in range(n_agents):
		agents.append(_generate_agent(cluster_id, list_home[n], list_work[n], list_geometry[n]))
	return agents



import numpy as np
import pandas as pd
from humobi.structures.trajectory import TrajectoriesFrame


def _Mseq(places,length,prob):
	out_seq = []
	for n in range(length):
		seq_type = np.random.choice([True,False], 1, p=[prob,1-prob])
		if seq_type:
			out_seq.append(n%places)
		else:
			out_seq.append(np.random.randint(0, places, 1)[0])
	return out_seq


def markovian_sequences_generator(users,places,length,prob):
	"""
	Generates synthetic frame wtih Markovian sequences.

	Args:
		users: The number of users (sequences) to generate.
		places: The number of unique places (symbols) to choose from. Can be a range or list to randomly choose from.
		length: The length of sequences. Can be a range or list to randomly choose from.
		prob: The $p$ value determining the probability of generator following the deterministic sequence.
		Can be a range or list to randomly choose from.

	Returns:
		A DataFrame with synthetic Markovian sequences.
	"""
	if isinstance(places,list):
		places = np.random.choice(places,users)
	else:
		places = [places] * users
	if isinstance(length,list):
		length = np.random.choice(length,users)
	else:
		length = [length] * users
	if isinstance(prob,list):
		prob = np.random.choice(prob,users)
	else:
		prob = [prob] * users
	frames = []
	for uid in range(users):
		generated_track = _Mseq(places[uid],length[uid],prob[uid])
		tmstmps = pd.date_range(0,periods=length[uid],freq='h')
		generated_frame = pd.concat({uid: pd.DataFrame(generated_track,index=tmstmps)})
		frames.append(generated_frame)
	generated_frame = pd.concat(frames)
	generated_frame = generated_frame.reset_index()
	generated_frame.columns = ['user_id', 'datetime', 'labels']
	generated_frame = generated_frame.set_index(['user_id','datetime'])
	return TrajectoriesFrame(generated_frame)


def random_sequences_generator(users,places,length):
	"""
	Generates synthetic frame wtih random sequences.

	Args:
		users: The number of users (sequences) to generate.
		places: The number of unique places (symbols) to choose from. Can be a range or list to randomly choose from.
		length: The length of sequences. Can be a range or list to randomly choose from.

	Returns:
		A DataFrame with synthetic random sequences.
	"""
	if isinstance(places,list):
		places = np.random.choice(places,users)
	else:
		places = [places] * users
	if isinstance(length,list):
		length = np.random.choice(length,users)
	else:
		length = [length] * users
	frames = []
	for uid in range(users):
		generated_track = np.random.randint(0,places[uid],length[uid])
		tmstmps = pd.date_range(0,periods=length[uid],freq='h')
		generated_frame = pd.concat({uid: pd.DataFrame(generated_track,index=tmstmps)})
		frames.append(generated_frame)
	generated_frame = pd.concat(frames)
	generated_frame = generated_frame.reset_index()
	generated_frame.columns = ['user_id', 'datetime', 'labels']
	generated_frame = generated_frame.set_index(['user_id','datetime'])
	return TrajectoriesFrame(generated_frame)


def deterministic_sequences_generator(users,places,repeats):
	"""
	Generates synthetic frame wtih deterministic sequences.

	Args:
		users: The number of users (sequences) to generate.
		places: The number of unique places (symbols) to choose from. Can be a range or list to randomly choose from.
		repeats: The number of repeats determining the lenght of sequences. Can be a range or list to randomly choose
	from.

	Returns:
		A DataFrame with synthetic deterministic sequences.
	"""
	if isinstance(places,list):
		places = np.random.choice(places,users)
	else:
		places = [places] * users
	if isinstance(repeats,list):
		repeats = np.random.choice(repeats,users)
	else:
		repeats = [repeats] * users
	frames = []
	for uid in range(users):
		generated_track = [x for x in range(places[uid])] * repeats[uid]
		tmstmps = pd.date_range(0,periods=len(generated_track),freq='h')
		generated_frame = pd.concat({uid: pd.DataFrame(generated_track,index=tmstmps)})
		frames.append(generated_frame)
	generated_frame = pd.concat(frames)
	generated_frame = generated_frame.reset_index()
	generated_frame.columns = ['user_id', 'datetime', 'labels']
	generated_frame = generated_frame.set_index(['user_id','datetime'])
	return TrajectoriesFrame(generated_frame)


def exploratory_sequences_generator(users, places):
	"""
	Generates synthetic frame wtih exploratory sequences, where each next symbol is previously unseen.

	Args:
		users: The number of users (sequences) to generate.
		places: The number of unique places (symbols) to choose from. Also determine the legnth of the sequence.
		Can be a range or list to randomly choose from.

	Returns:
		A DataFrame with synthetic exploratory sequences.
	"""
	if isinstance(places,list):
		places = np.random.choice(places,users)
	else:
		places = [places] * users
	frames = []
	for uid in range(users):
		generated_track = np.random.permutation([x for x in range(places[uid])])
		tmstmps = pd.date_range(0,periods=len(generated_track),freq='h')
		generated_frame = pd.concat({uid: pd.DataFrame(generated_track,index=tmstmps)})
		frames.append(generated_frame)
	generated_frame = pd.concat(frames)
	generated_frame = generated_frame.reset_index()
	generated_frame.columns = ['user_id', 'datetime', 'labels']
	generated_frame = generated_frame.set_index(['user_id','datetime'])
	return TrajectoriesFrame(generated_frame)


def self_transitions_sequences_generator(users,places,length):
	"""
	Generates synthetic frame wtih self-transitions sequences. The number of self-transtions repeating after each other 
	is determined by the number of symbols and the legnth of the sequence.

	Args:
		users: The number of users (sequences) to generate.
		places: The number of unique places (symbols) to choose from. Can be a range or list to randomly choose from.
		length: The length of sequences. Can be a range or list to randomly choose from.

	Returns:
		A DataFrame with synthetic self-transitions sequences.
	"""
	if isinstance(places,list):
		places = np.random.choice(places,users)
	else:
		places = [places] * users
	if isinstance(length,list):
		length = np.random.choice(length,users)
	else:
		length = [length] * users
	frames = []
	for uid in range(users):
		per_place = length[uid]//places[uid]
		generated_track = []
		for n in range(places[uid]):
			generated_track += [n]*per_place
		tmstmps = pd.date_range(0,periods=len(generated_track),freq='h')
		generated_frame = pd.concat({uid: pd.DataFrame(generated_track,index=tmstmps)})
		frames.append(generated_frame)
	generated_frame = pd.concat(frames)
	generated_frame = generated_frame.reset_index()
	generated_frame.columns = ['user_id', 'datetime', 'labels']
	generated_frame = generated_frame.set_index(['user_id','datetime'])
	return TrajectoriesFrame(generated_frame)


def non_stationary_sequences_generator(users, places, states, length):
	"""
	Generates synthetic frame wtih non-stationary sequences. The algorithms chooses between states, each with different
	generation routine.

	Args:
		users: The number of users (sequences) to generate.
		places: The number of unique places (symbols) to choose from. Can be a range or list to randomly choose from.
		states: The number of states with different generation routine. Can be a range or list to randomly choose
		from.
		length: The length of sequences. Can be a range or list to randomly choose from.

	Returns:
		A DataFrame with synthetic non-stationary sequences.
	"""
	if isinstance(places,list):
		places = np.random.choice(places,users)
	else:
		places = [places] * users
	if isinstance(length,list):
		length = np.random.choice(length,users)
	else:
		length = [length] * users
	if isinstance(states,list):
		states = np.random.choice(states,users)
	else:
		states = [states] * users
	frames = []
	for uid in range(users):
		num_of_states = states[uid]
		states_probs = np.random.uniform(0, 1, num_of_states)
		states_probs /= sum(states_probs)
		states_steps = np.random.choice([x for x in range(num_of_states)],length[uid],p=states_probs)
		states_places = [np.random.uniform(0, 1, places[uid]) for x in range(num_of_states)]
		states_places = [x/sum(x) for x in states_places]
		generated_track = []
		for n in range(length[uid]):
			generated_track.append(np.random.choice([x for x in range(places[uid])], p=states_places[states_steps[n]]))
		tmstmps = pd.date_range(0, periods=len(generated_track), freq='h')
		generated_frame = pd.concat({uid: pd.DataFrame(generated_track, index=tmstmps)})
		frames.append(generated_frame)
	generated_frame = pd.concat(frames)
	generated_frame = generated_frame.reset_index()
	generated_frame.columns = ['user_id', 'datetime', 'labels']
	generated_frame = generated_frame.set_index(['user_id','datetime'])
	return TrajectoriesFrame(generated_frame)



from ...structures.trajectory import TrajectoriesFrame
import sys
sys.path.append("..")
from humobi.models.temporal_tools import cluster_traj
from humobi.models.spatial_tools import misc, distributions, generating, filtering
from humobi.misc import create_grid
from humobi.models.spatial_tools.misc import rank_freq
from math import ceil
from humobi.models.agent_module.generate_agents import generate_agents


WEIGHT = False
SIDE = 1000


path = "C:\\Users\\ppp\\Desktop\\staz\\repo-HuMobi2\\HuMobi2\\oryginal.csv"
trajectories_frame = TrajectoriesFrame(path, {
	'names': ['id', 'datetime', 'temp', 'lat', 'lon', 'labels', 'start', 'end', 'geometry'], 'crs': "EPSG:27700",
	'delimiter': ',', 'skiprows': 1, 'nrows': 9386})

circadian_collection, cluster_association, cluster_share = cluster_traj.cluster_trajectories(trajectories_frame, weights=WEIGHT)
commute_dist = distributions.commute_distances(trajectories_frame, quantity = 2)
layer = create_grid.create_grid(trajectories_frame, resolution = SIDE)
layer = filtering.filter_layer(layer,trajectories_frame)
unique_labels = set(cluster_association.values()).difference(set([-1]))
sig_frame = rank_freq(trajectories_frame, quantity = 2)
cluster_spatial_distributions = {}
cluster_commute_distributions = {}
for n in [unique_labels][0]:
	group_indicies = [k for k,v in cluster_association.items() if v == n]
	group_sig_frame = sig_frame.loc[group_indicies]
	group_commute_dist = {k:v.loc[group_indicies] for k,v in commute_dist.items()}
	dist_list = distributions.convert_to_2d_distribution(sig_frame, layer, crs="epsg:27700", return_centroids=True, quantity = 2)
	commute_distributions = distributions.commute_distances_to_2d_distribution(group_commute_dist, layer, crs="epsg:27700", return_centroids=True)
	cluster_spatial_distributions[n] = dist_list
	cluster_commute_distributions[n] = commute_distributions
to_generate = 29
generated_agents = []
for label, share in cluster_share.items():
	amount = ceil(share*to_generate)
	current_spatial_distributions = cluster_spatial_distributions[label]
	current_commute_distributions = cluster_commute_distributions[label]
	home_positions = generating.generate_points_from_distribution(current_spatial_distributions[0], amount)
	work_positions = generating.select_points_with_commuting(home_positions,current_spatial_distributions[1],current_commute_distributions)
	activity_areas = generating.generate_activity_areas('ellipse', home_positions, work_positions, layer, 1.0)
	agents = generate_agents(amount, label, home_positions, work_positions, activity_areas)
	generated_agents += agents
circadian_rhythm = cluster_traj.circadian_rhythm_extraction(circadian_collection)


import numpy as np
import pandas as pd
import geopandas as gpd
from src.humobi.models.spatial_tools.misc import normalize_array
import matplotlib.pyplot as plt
from shapely.geometry import Point, Polygon
import math


def generate_points_from_distribution(distribution, amount):
	"""
	Randomly chooses points from given distributions
	:param distribution: The list with distribution
	:param amount: The number of points to choose
	:return: A vector of chosen points
	"""
	choices = []
	probabilities = distribution.counted.values.astype('float')
	for n in range(amount):
		choose_from = normalize_array(probabilities)
		choice = np.random.choice(distribution.index.values, p=choose_from)
		choices.append(distribution.loc[choice].geometry)
		wherewasit = np.argwhere(distribution.index.values == choice)[0][0]
		probabilities[wherewasit] -= 1/amount
		probabilities[probabilities < 0] = 0
	return choices


def select_points_with_commuting(starting_positions, target_distribution, commuting, spread=None):
	"""
	Returns an array with work positions
	:param starting_positions: contains home positions
	:param target_distribution: contains workplace distribution
	:param commuting: contains commuting distance distribution by each cell
	:param spread: an optional parameter, the value of spread to create an annulus
	:return: an array with work positions
	"""
	target_distribution = target_distribution[['geometry', 'counted']].copy()
	distance_indicies = [indeks[0] for indeks in [
		commuting[1][commuting[1]['geometry'] == home].index.values for home in starting_positions]]
	distances = commuting[1].loc[distance_indicies]
	if spread is None:
		buffers = [starting_positions[dist].buffer(distances.values[dist][0]) * 1.01 for dist in range(len(distances))]
	else:
		buffers = [starting_positions[dist].buffer(distances.values[dist][0] * (1.0 + spread)).difference(
			starting_positions[dist].buffer(distances.values[dist][0] * (1.0 - spread))) for dist in
			range(len(distances))]
	chosen_work_places = []
	index = 0
	for buffer in buffers:
		if buffer.area == 0.0:
			chosen = None
		# chosen = gpd.GeoDataFrame(pd.DataFrame(starting_positions), geometry = 0).loc[index][0]  # decide on the option
		else:
			inside_buffer = target_distribution.loc[target_distribution['geometry'].intersects(buffer) == True]
			if len(inside_buffer) != 0 and not np.all(inside_buffer.values[:, 1] == 0):
				probabilities = normalize_array(inside_buffer.values[:, 1].astype('float'))
				chosen = np.random.choice(inside_buffer.values[:, 0], size=1, p=probabilities)[0]
				where = target_distribution[target_distribution.geometry == chosen].index
				target_distribution.loc[where,'counted'] = target_distribution.loc[where,'counted'] - 1/len(buffers)
				target_distribution.loc[target_distribution.counted < 0,'counted'] = 0
			else:
				chosen = None
		chosen_work_places.append(chosen)
		index += 1
	return np.array(chosen_work_places)


def create_ellipse(home_location, work_location, spread):
	"""
	Returns ellipse with a home at its centre and work location at the edge
	:param home_location: contains a home position;
	:param work_location: contains a work position;
	:param spread: contains a ratio between major and minor axis;
	:return: ellipse with a home at its centre and work location at the edge.
	"""
	if work_location is None:
		return None
	else:
		a = home_location.distance(work_location)
		b = a * float(spread)
		point_list = []
		azimuth = math.atan2(work_location.y - home_location.y, work_location.x - home_location.x)
		ro = (math.pi / 200)

		for t in range(0, 401):
			x = home_location.x + (a * math.cos(t * ro) * math.cos(azimuth) - b * math.sin(t * ro) * math.sin(azimuth))
			y = home_location.y + (b * math.sin(t * ro) * math.cos(azimuth) + a * math.cos(t * ro) * math.sin(azimuth))
			point_list.append([Point(x, y).x, Point(x, y).y])
		return Polygon(point_list)


def generate_activity_areas(area_type, home_positions, work_positions, layer, spread):
	"""
	Returns an array with acitvity areas
	:param area_type: the type of activity area
	:param home_positions: contains home locations
	:param work_positions: contains work locations
	:param spread: contains a ratio between major and minor axis
	:return: an array with acitvity areas
	"""
	activity_areas = []
	if area_type == 'ellipse':
		for index in range(0, len(home_positions)):
			ellipse = create_ellipse(home_positions[index], work_positions[index], spread)
			try:
				intersection_layer = dict(layer.intersects(ellipse))
				gridded_activity_area = {k: v for k, v in intersection_layer.items() if v == True}
				activity_area = [layer.loc[n] for n in list(gridded_activity_area.keys())]
				activity_areas.append(activity_area)
			except:
				activity_areas.append(None)
		return np.array(activity_areas)  # return activity_areas
	else:
		pass



import numpy as np
import pandas as pd
from scipy.optimize import fsolve
from numba import cuda, jit
from math import ceil
from Bio import pairwise2

def normalize_chain(dicto):
	"""
	Normalizes dictionary values. Used for the Markov Chain normalization.

	Args:
		dicto: dictionary to ..

	Returns:
		..d dictionary
	"""
	total = 1 / float(np.sum(list(dicto.values())))
	for k, v in dicto.items():
		dicto[k] = v * total
	return dicto


def get_diags(a):
	"""
	Extracts all the diagonals from the matrix.

	Args:
		a: numpy array to process

	Returns:
		a list of diagonals
	"""
	diags = [a.diagonal(i) for i in range(-a.shape[0] + 1, a.shape[1])]
	return [n.tolist() for n in diags]


def resolution_to_points(r, c_max, c_min):
	"""
	Calculates how many points are needed to divide range for given resolution
	:param r: resolution
	:param c_max: maximum value
	:param c_min: minimum value
	:return: the number of points
	"""
	c_range = c_max - c_min
	c_points = c_range / r + 1
	return ceil(c_points)


def moving_average(a, n=2):
	"""
	Implements fast moving average
	:param a: input data array
	:param n: window size
	:return: processed data array
	"""
	ret = np.cumsum(a, dtype=float)
	ret[n:] = ret[n:] - ret[:-n]
	return ret[n - 1:] / n


def _matchfinder(start_idx, gs, data_len):
	"""
	Finds the shortest not repeating sequences according to the Lempel-Ziv algorithm
	:param start_idx: starting point in the array from which search will be started
	:param gs: symbol series
	:param data_len: data length
	:return: current starting index and the shortest non-repeating subsequence length
	"""
	max_subsequence_matched = 0
	for i in range(0, start_idx):
		j = 0
		end_distance = data_len - start_idx
		while (start_idx + j < data_len) and (i + j < start_idx) and (gs[i + j] == gs[start_idx + j]):
			j += 1
		if j == end_distance:
			return start_idx, 0
		elif j > max_subsequence_matched:
			max_subsequence_matched = j
	return start_idx, max_subsequence_matched + 1


@cuda.jit
def _matchfinder_gpu(gs, data_len, output):
	"""
	Finds the shortest not repeating sequences according to the Lempel-Ziv algorithm. Algorithm adaptation for GPU.
	:param gs: symbol series
	:param data_len: data length
	:param output: output array
	"""
	pos = cuda.grid(1)
	max_subsequence_matched = 0
	finish_bool = False
	if pos < data_len:
		for i in range(0, pos):
			j = 0
			end_distance = data_len - pos
			while (pos + j < data_len) and (i + j < pos) and (gs[i + j] == gs[pos + j]):
				j += 1
			if j == end_distance:
				finish_bool = True
				break
			elif j > max_subsequence_matched:
				max_subsequence_matched = j
		if finish_bool:
			output[pos] = end_distance + 1  # CHANGED with XU paper
		else:
			output[pos] = max_subsequence_matched + 1


def matchfinder(gs, gpu=True):
	"""
	Finds the shortest not repeating sequences according to the Lempel-Ziv algorithm
	:param gs: symbol series
	:return: the length of the shortest non-repeating subsequences at each step of sequence
	"""
	gs = gs.dropna()
	data_len = len(gs)
	gs = np.array(gs.values)
	output = np.zeros(data_len)
	output[0] = 1
	if gpu:
		threadsperblock = 256
		blockspergrid = ceil(data_len / threadsperblock)
		_matchfinder_gpu[threadsperblock, blockspergrid](gs, data_len, output)
	return output


@jit
def _repeatfinder_dense(s1, s2):
	output = np.zeros(len(s2))
	for pos1 in range(0, len(s2)):
		max_s = 0
		for pos2 in range(0, len(s1)):
			j = 0
			while s1[pos2 + j] == s2[pos1 + j]:
				j += 1
				if pos2 + j == len(s1) or pos1 + j == len(s2):
					break
			if j > max_s:
				max_s = j
		output[pos1] = max_s
	return max(output)-1 / len(s2)-1


@jit
def _repeatfinder_sparse(s1, s2):
	matrix = [[0 for x in range(len(s2))] for x in range(len(s1))]
	for i in range(len(s1)):
		for j in range(len(s2)):
			if s1[i] == s2[j]:
				if i == 0 or j == 0:
					matrix[i][j] += 1
				else:
					matrix[i][j] = matrix[i - 1][j - 1] + 1
			else:
				matrix[i][j] = max(matrix[i - 1][j], matrix[i][j - 1])
	cs = matrix[-1][-1]
	return cs-1 / len(s2)-1


def _repeatfinder_equally_sparse(s1, s2):
	matrix = np.zeros((len(s1), len(s2)))  # prepare matrix for results
	for i in range(len(s1)):
		for j in range(len(s2)):
			if s1[i] == s2[j]:
				if i == 0 or j == 0:  # if matched symbols are at the start of the sequence
					matrix[i][j] += 1  # if symbols matched - add 1
				else:
					matrix[i][j] = 1  # if symbols matched - add 1
	s2_indi = (np.vstack(
		[np.arange(matrix.shape[1]) for x in range(matrix.shape[0])]) + 1) * matrix  # convert matched 1's into indices
	s2diags = get_diags(s2_indi)  # get all diagonals
	nonzero_transitions = [[y for y in x if y != 0] for x in s2diags]  # filter out empty lists
	nonzero_transitions = [[(x[y], x[y + 1]) for y in range(len(x) - 1)] for x in
	                       nonzero_transitions]  # convert to transitions
	translateds2 = [[(s2[int(y[0]) - 1], s2[int(y[1]) - 1]) for y in x] for x in nonzero_transitions if
	                len(x) > 0]  # convert to matched symbols
	# search for self-transitions and transitions
	try:
		selft = sum([x[0] == x[1] for x in sorted(translateds2, key=len)[-1]])
	except:
		selft = 0
	try:
		nonselft = sum([x[0] != x[1] for x in sorted(translateds2, key=len)[-1]])
	except:
		nonselft = 0
	return sum([selft, nonselft]) / (len(s2) - 1)


def _global_align(s1, s2):
	one = list(s1)
	two = list(s2)
	alignment = \
		pairwise2.align.globalms(one, two, 1, -1, -1, 0, penalize_end_gaps=False, one_alignment_only=True,
		                         gap_char=['-'])[
			0]
	return alignment.score / (len(two)-1)


def _iterative_global_align(s1, s2):
	one = list(s1)  # preapare lists of symbols
	two = list(s2)
	cut = two  # assign currently processed sequence
	all_match = []
	to_search = []
	while True:
		best_match = \
			pairwise2.align.globalms(one, cut, 1, -1, -1, 0, penalize_end_gaps=False,
			                         one_alignment_only=True, gap_char=['-'])[
				0]
		zipped = [(x, y) for x, y in zip(best_match[0], best_match[1])]  # combinations of matched symbols
		out_of_match = [1 if x[0] != x[1] and x[0] == '-' else 0 for x in zipped]  # search for mismatched symbols
		out_diff = np.diff(out_of_match)  # find gaps
		starts = [x for x in range(len(out_diff)) if out_diff[x] == 1]  # find starts of gaps
		ends = [x for x in range(len(out_diff)) if out_diff[x] == -1]  # find ends of gaps
		lengths = [y - x for x, y in zip(starts, ends)]  # find lengths of gaps
		first_ends = [1 if x == '-' else 0 for x in best_match[0]]  # find mismatched starts and ends of sequence
		first_ends_diff = np.diff(first_ends)  # find range of mismatched starts and ends
		first_ends_starts = [x for x in range(len(first_ends_diff)) if first_ends_diff[x] == 1]
		first_ends_ends = [x for x in range(len(first_ends_diff)) if first_ends_diff[x] == -1]
		if len(first_ends_ends) > 0 and 0 not in first_ends[:first_ends_ends[0]]:
			begin = best_match[1][:first_ends_ends[0] + 1]
		else:
			begin = []
		if len(first_ends_starts) > 0 and 0 not in first_ends[first_ends_starts[-1] + 1:]:
			end = best_match[1][first_ends_starts[-1] + 1:]
		else:
			end = []
		lengths += [len(begin), len(end)]
		maxleng = np.max(lengths)  # check which mismatch is the longest
		longest = np.where(lengths == maxleng)  # take the mismatched part
		if longest[0][0] == len(lengths) - 1:  # check if its a gap in the middle or at the ends
			to_search.append(end)
		elif longest[0][0] == len(lengths) - 2:
			to_search.append(begin)
		else:
			for n in longest[0]:
				n = n
				if n == len(lengths) - 1:
					to_search.append(end)
				elif n == len(lengths) - 2:
					to_search.append(begin)
				else:
					to_search.append(best_match[1][starts[n] + 1:ends[n] + 1])
		if len(to_search) == 0:  # if there are no sequences to match - stop
			if best_match[2] > 0:  # if there was positive score in the last match - add it to the list
				all_match.append(best_match[2] - 1)  # add score to the list of scores (-1 for transitions)
			break
		cut = to_search.pop(0)  # pop out the sequence for search
		cut = [x for x in cut if isinstance(x, float)]  # take only symbols
		if best_match[2] <= 0:  # if there is zero score already - stop
			break
		all_match.append(best_match[2] - 1)  # add score to the list of scores (-1 for transitions)
		if len(cut) <= 1:
			break  # if sequence for search does not consist of at least two symbols - stop
	return sum(all_match) / (len(two) - 1)


def _equally_sparse_match(s1, s2):
	if len(s1) > len(s2):
		return None
	matrix = np.zeros((len(s1), len(s2)))  # prepare matrix for results
	for i in range(len(s1)):
		for j in range(len(s2)):
			if s1[i] == s2[j]:
				if i == 0 or j == 0:  # if matched symbols are at the start of the sequence
					matrix[i][j] += 1  # if symbols matched - add 1
				else:
					matrix[i][j] = 1  # if symbols matched - add 1
	s2_indi = (np.vstack(
		[np.arange(matrix.shape[1]) + 1 for x in range(matrix.shape[0])])) * matrix  # convert matched 1's into indices
	s1_indi = (np.hstack(
		[np.expand_dims(np.arange(matrix.shape[0]), axis=1) + 1 for x in range(matrix.shape[1])])) * matrix
	s2diags = get_diags(s2_indi)  # get all diagonals
	s1diags = get_diags(s1_indi)
	if sum([sum(x) for x in s2diags]) == 0:
		return None
	nonzero_s2 = [[y - 1 for y in x if y != 0] for x in s2diags if sum(x) > 0]  # filter out empty lists
	nonzero_s1 = [[len(s1) - y + 1 for y in x if y != 0] for x in s1diags if sum(x) > 0]  # filter out empty lists
	# nonzero_s2 = [x for x in nonzero_s2 if len(x) >= 2]
	# nonzero_s1 = [x for x in nonzero_s1 if len(x) >= 2]
	matches = []
	for x, y in zip(nonzero_s1, nonzero_s2):
		if y[-1] + x[-1] < len(s2):
			matched_pattern = [(int(-w), s2[int(z)]) for z, w in zip(y, x)]
			next_symbol = s2[int(y[-1] + x[-1])]
			matches.append((matched_pattern, next_symbol))
	return matches


def fano_inequality(distinct_locations, entropy):
	"""
	Implementation of the Fano's inequality. Algorithm solves it and returns the solution.
	:param distinct_locations:
	:param entropy:
	:return:
	"""
	func = lambda x: (-(x * np.log2(x) + (1 - x) * np.log2(1 - x)) + (1 - x) * np.log2(
		distinct_locations - 1)) - entropy
	return fsolve(func, .9999)[0]


def to_labels(trajectories_frame):
	"""
	Adds labels column based on repeating geometries or coordinates
	:param trajectories_frame: TrajectoriesFrame object class
	:return: TrajectoriesFrame with labels column
	"""
	to_tranformation = trajectories_frame[trajectories_frame.geometry.is_valid]
	try:
		to_tranformation['labels'] = to_tranformation[to_tranformation._geom_cols[0]].astype(str) + ',' + \
		                               to_tranformation[to_tranformation._geom_cols[1]].astype(str)
		trajectories_frame['labels'] = trajectories_frame[to_tranformation._geom_cols[0]].astype(str) + ',' + \
		                             trajectories_frame[to_tranformation._geom_cols[1]].astype(str)
	except:
		to_tranformation['labels'] = to_tranformation.lat.astype(str) + ',' + to_tranformation.lon.astype(str)
		trajectories_frame['labels'] = trajectories_frame.lat.astype(str) + ',' + trajectories_frame.lon.astype(str)
	unique_coors = pd.DataFrame(pd.unique(to_tranformation['labels']))
	unique_coors.index = unique_coors.loc[:, 0]
	unique_coors.loc[:, 0] = range(len(unique_coors))
	sub_dict = unique_coors.to_dict()[0]
	trajectories_frame.astype({'labels': str})
	trajectories_frame['labels'] = trajectories_frame['labels'].map(sub_dict)
	return trajectories_frame



import numpy as np
from tqdm import tqdm

tqdm.pandas()


def normalize_chain(dicto):
	"""
	Normalizes dictionary values. Used for the Markov Chain normalization.
	:param dicto: dictionary to normalize
	:return: normalized dictionary
	"""
	total = 1 / float(np.sum(list(dicto.values())))
	for k, v in dicto.items():
		dicto[k] = v * total
	return dicto


def build_single(sequence, state_size, state):
	"""
	A small temporary model for lower order Markov Chains called during prediction when previously unseen state is
	encountered.
	:param sequence: Sequence to learn
	:param state_size: Order of the Markov Chain
	:param state: Current state
	:return: Lower order Markov Chain
	"""
	model = {}
	for posnum in range(len(sequence) - state_size):
		if tuple(sequence[posnum:posnum + state_size]) == state:
			next = sequence[posnum + state_size]
			if state not in model.keys():
				model[state] = {}
			if next not in model[state].keys():
				model[state][next] = 0
			model[state][next] += 1
	return model


class MarkovChain(object):
	"""
	Markov Chain class.
	:param sequence: Sequence which is used for chain building
	:param state_size: The order of the Markov Chain
	"""

	def __init__(self, sequence, state_size):
		"""
		Class initialization. Calls chain building (learning).
		"""
		self._state_size = state_size
		self._sequence = sequence
		self.model = self.build()

	@property
	def state_size(self):
		return self._state_size

	@property
	def sequence(self):
		return self._sequence

	@sequence.setter
	def sequence(self, value):
		self._sequence = value

	def build(self):
		"""
		Builds the Markov Chain. Returned model is not normalized as it is normalized during prediction.
		:return: Markov Model
		"""
		model = {}  # the model is a dictionary
		for posnum in range(len(self.sequence) - self.state_size):  # for each element in the sequence
			state = tuple(self.sequence[posnum:posnum + self.state_size])  # read current state, including order
			next = self.sequence[posnum + self.state_size]  # read the next symbol
			if state not in model.keys():  # if symbol not yet encountered
				model[state] = {}  # create a slot for it
			if next not in model[state].keys():  # if symbol encountered but the next state haven't been encountered yet for that symbol
				model[state][next] = 0  # create a slot for it
			model[state][next] += 1  # count +1 for that transition
		return model

	def move(self, state):
		"""
		Predict the next symbol based on the given state.
		:param state: The state from which the Markov Chain will make prediction
		:return: Predicted symbol
		"""
		state = tuple(state)  # current state - the model operates on tuples
		if state not in self.model.keys():  # this whole if operates on the case when the state is not met and fits lower-order models
			for lower in range(1, self.state_size + 1):
				lower_state = state[lower::]
				lower_state_size = self.state_size - lower
				lower_state_sequence = [tuple(self.sequence[posnum:posnum + lower_state_size]) \
				                        for posnum in range(len(self.sequence) - lower_state_size)]
				if lower_state in lower_state_sequence:
					break
				if len(lower_state) == 0:
					return -1
			temp_model = build_single(self.sequence, lower_state_size, lower_state)  # builds temporal, smaller model
			transit = normalize_chain(temp_model[lower_state])
		else:
			transit = normalize_chain(self.model[state])  # normalize chain
		prediction = np.random.choice(list(transit.keys()), p=list(transit.values()))  # make prediction
		return prediction

	def move_from_build(self, horizon, update=False):
		"""
		Predict the next symbols based on the last seen state during training
		:param horizon: How many symbols to predict
		:param update: Whether the model should be updated by the new predictions (not recommended)
		:return: Predicted symbols
		"""
		predicted_sequence = []
		recent_state = self.sequence[-self.state_size:]
		horizon -= self.state_size
		for steps in range(horizon):
			prediction = self.move(recent_state)
			predicted_sequence.append(prediction)
			recent_state.append(prediction)
			recent_state = recent_state[1:]
			if update:
				self.sequence += prediction
				self.build()
		return np.array(predicted_sequence)



import pandas as pd
from humobi.misc.utils import to_labels


def normalize_array(arr):
	"""
	Normalizes given array (sum to 1)
	:param arr: Array to normalize
	:return: Normalized array
	"""
	arr = arr/arr.sum()
	if arr.sum() != 1:
		offset = 1 - arr.sum()
		arr[arr.argmax()] += offset
	return arr


def rank_freq(trajectories_frame, quantity=2):
	"""
	Ranks locations visited by each user by the frequency of visits. Used to select the most important location
	:param trajectories_frame: TrajectoriesFrame class object
	:param quantity: The number of location to rank
	:return: Ranked locations
	"""
	if not 'labels' in trajectories_frame:
		to_labels(trajectories_frame)
	filtered_trajectories_frame = trajectories_frame[trajectories_frame.geometry.is_valid]
	top_places = filtered_trajectories_frame.groupby(
		[filtered_trajectories_frame.index.get_level_values(0), 'labels']).count().reset_index()
	top_places = pd.DataFrame(top_places).groupby('user_id').apply(lambda x: x.sort_values('temp', ascending=False)).reset_index(drop=True)
	merged = pd.merge(top_places[['user_id', 'labels']], filtered_trajectories_frame.reset_index(), on=['user_id', 'labels'], how='left')[['user_id', 'labels', 'geometry']].drop_duplicates()
	if quantity == 1:
		unstacked = merged.groupby('user_id')['geometry'].apply(lambda x: x.reset_index(drop=True)).unstack().iloc[:,quantity - 1]
	else:
		unstacked = merged.groupby('user_id')['geometry'].apply(lambda x: x.reset_index(drop=True)).unstack().iloc[:,:quantity]
	return unstacked



import numpy as np
from shapely.geometry import Point
from humobi.misc.utils import resolution_to_points, moving_average, to_labels
from tqdm import tqdm
import pandas as pd
import geopandas as gpd
import concurrent.futures as cf
from itertools import repeat
from humobi.structures.trajectory import TrajectoriesFrame


class GridAggregation():
	"""
	A class for spatial aggregation of movement trajectories. This one uses a regular grid.
	"""

	def __init__(self, resolution=300, origin=True, x_min=None, x_max=None, y_min=None, y_max=None):
		"""
		Class initialisation. Depending on what is defined, the grid fits to the given data spatial extent of in the middle
		of data.

		Args:
			resolution (default = 300): grid resolution
			origin: Determines if the grid should be centered to the centroid of the data
			x_min: minimum x value of the grid
			x_max: maximum x value of the grid
			y_min: minimum y value of the grid
			y_max: maximum y value of the grid
		"""
		self._x_min = x_min
		self._x_max = x_max
		self._y_min = y_min
		self._y_max = y_max

		self._resolution = resolution
		if isinstance(origin, bool):
			self._detect_origin = True
		else:
			self._detect_origin = False
		if x_min is None or x_max is None or y_min is None or y_max is None:
			self._detect_range = True
		else:
			self._detect_range = False
		if not self._detect_origin and not self._detect_range:
			raise AttributeError("Cannot set origin and range at the same time!")

	@property
	def resolution(self):
		return self._resolution

	@property
	def x_min(self):
		return self._x_min

	@property
	def x_max(self):
		return self._x_max

	@property
	def y_min(self):
		return self._y_min

	@property
	def y_max(self):
		return self._y_max

	def _aggregate_multi(self, x, y, centroid_x, centroid_y, row_index, val):
		"""
		Fast spatial aggregation of movement trajectories spatial aggregation. Multithread variant.

		Args:
			x: grid x coordinates
			y: grid y coordinates
			centroid_x: grid x coordinates of centroids
			centroid_y: grid y coordinates of centroids
			row_index: data row index
			val: data row value

		Returns:
			assigned geometry and row index
		"""
		lower_x = np.max(np.where((val.x >= x)))
		lower_y = np.max(np.where((val.y >= y)))
		if lower_x == len(centroid_x):
			assigned_x = centroid_x[lower_x - 1]
		else:
			assigned_x = centroid_x[lower_x]
		if lower_y == len(centroid_y):
			assigned_y = centroid_y[lower_y - 1]
		else:
			assigned_y = centroid_y[lower_y]
		return row_index, Point(assigned_x, assigned_y)

	def _aggregate(self, x, y, centroid_x, centroid_y, val):
		"""
		Fast movement trajectories spatial aggregation
		Args:
			x: grid x coordinates
			y: grid y coordinates
			centroid_x: grid x coordinates of centroids
			centroid_y: grid y coordinates of centroids
			row_index: data row index
			val: data row value

		Returns:
			assigned geometry and row index
		"""
		lower_x = np.max(np.where((val.x >= x)))
		lower_y = np.max(np.where((val.y >= y)))
		if lower_x == len(centroid_x):
			assigned_x = centroid_x[lower_x - 1]
		else:
			assigned_x = centroid_x[lower_x]
		if lower_y == len(centroid_y):
			assigned_y = centroid_y[lower_y - 1]
		else:
			assigned_y = centroid_y[lower_y]
		return assigned_x, assigned_y

	def aggregate(self, trajectories_frame, inplace=True, parallel=False):
		"""
		Aggregates TrajectoriesFrame object spatially.

		Args:
			trajectories_frame: TrajectoriesFrame class object
			inplace: determines if coordinates should be overwritten. If false, a new column is added.
			parallel: determines if should uses multithreading (default = False, because it is slower for small data)

		Returns:
			A TrajectoriesFrame with aggregated geometry.
		"""
		trajectories_copy = trajectories_frame.copy()
		if self._detect_range and self._detect_origin:
			self._x_min = trajectories_frame['geometry'].x.min()
			self._x_max = trajectories_frame['geometry'].x.max()
			self._y_min = trajectories_frame['geometry'].y.min()
			self._y_max = trajectories_frame['geometry'].y.max()
		x_points = resolution_to_points(self.resolution, self.x_max, self.x_min)
		y_points = resolution_to_points(self.resolution, self.y_max, self.y_min)
		x = np.linspace(self.x_min, self.x_max, x_points)
		y = np.linspace(self.y_min, self.y_max, y_points)
		centroid_x = moving_average(x)
		centroid_y = moving_average(y)
		agg_dict = {}
		if parallel:
			with cf.ThreadPoolExecutor(max_workers=6) as executor:
				indis = [indi for indi, val in trajectories_frame['geometry'].iteritems()]
				vals = [val for indi, val in trajectories_frame['geometry'].iteritems()]
				results = list(tqdm(
					executor.map(self._aggregate_multi, repeat(x), repeat(y), repeat(centroid_x), repeat(centroid_y),
					             indis, vals), total=trajectories_frame.shape[0]))
			for result in results:
				agg_dict[result[0]] = result[1]
		else:
			for indi, val in tqdm(trajectories_frame['geometry'].iteritems(), total=trajectories_frame.shape[0]):
				if np.isnan(val.x):
					agg_dict[indi] = Point(float('nan'), float('nan'))
					continue
				assigned_x, assigned_y = self._aggregate(x, y, centroid_x, centroid_y, val)
				agg_dict[indi] = Point(assigned_x, assigned_y)
		if inplace:
			trajectories_copy['geometry'].update(pd.Series(agg_dict))
			trajectories_copy[trajectories_copy._geom_cols[0]] = trajectories_copy["geometry"].x
			trajectories_copy[trajectories_copy._geom_cols[1]] = trajectories_copy["geometry"].y
		else:
			trajectories_copy['old_geometry'] = trajectories_copy['geometry']
			trajectories_copy['geometry'] = pd.Series(agg_dict)
			trajectories_copy[trajectories_copy._geom_cols[0]] = trajectories_copy["geometry"].x
			trajectories_copy[trajectories_copy._geom_cols[1]] = trajectories_copy["geometry"].y
			trajectories_copy['geometry'] = trajectories_copy['old_geometry']
		trajectories_copy = to_labels(trajectories_copy)
		return trajectories_copy


class ClusteringAggregator():
	"""
	A class for spatial aggregation of movement trajectories. This one uses clustering algorithms from sklearn library.
	"""

	def __init__(self, algorithm, **kwargs):
		"""
		Class initialisation. Accepts sklearn clustering algorithms' classes and their keyword arguments.

		Args:
			algorithm: Clustering algorithm from sklearn library.
			**kwargs: Any accepted kwargs.
		"""
		self._algorithm = algorithm(**kwargs)

	@property
	def algorithm(self):
		return self._algorithm

	def _user_aggregate(self, single_trajectory):
		"""
		Spatially aggregates single movement trajectory. Adds labels columns.

		Args:
			single_trajectory: Single movement trajectory

		Returns:
			An aggregated TrajectoriesFrame with labels column added.
		"""
		try:
			fited = self.algorithm.fit(single_trajectory)
		except:
			raise ValueError
		single_trajectory['labels'] = fited.labels_
		return single_trajectory

	def _recalcuate_centres(self, single_trajectory):
		"""
		Based on labels, recalculates spatial coordinates of points to their clusters centers.

		Args:
			single_trajectory: single movement trajectory

		Returns:
			a TrajectoriesFrame with overwritten coordinates in lon and lat columns
		"""
		centres = single_trajectory.groupby(by='labels').apply(lambda x: x.mean())
		single_trajectory = single_trajectory.join(centres, on='labels', rsuffix='_')
		single_trajectory['lat'] = single_trajectory['lat_']
		single_trajectory['lon'] = single_trajectory['lon_']
		return single_trajectory

	def aggregate(self, trajectories_frame, drop_noise=False, centres_as_geometry=True):
		"""
		General function for spatial data aggregation. It assigns labels to clusters and has additional options of
		centers recalculation and data overwriting.

		Args:
			trajectories_frame: TrajectoriesFrame object class
			drop_noise: Datermines if noise (-1 labels) should be drop
			centres_as_geometry: If true, centres of clusters are assigned as geometry
		Raturns:
			A spatially aggregated TrajectoriesFrame
		"""
		if not hasattr(trajectories_frame, '_geom_cols'):
			trajectories_frame = TrajectoriesFrame(trajectories_frame)
		coordinates_frame = trajectories_frame[[trajectories_frame._geom_cols[0], trajectories_frame._geom_cols[1]]]
		clustered = coordinates_frame.groupby(level=0).progress_apply(lambda x: self._user_aggregate(x))
		if drop_noise:
			clustered[clustered['labels'] == -1] = None
			if len(clustered) == 0:
				return None
		if centres_as_geometry:
			clustered = clustered.groupby(level=0).progress_apply(lambda x: self._recalcuate_centres(x))
			clustered['geometry'] = [Point(*x) for x in
			                         zip(clustered.loc[:, 'lat'],
			                             clustered.loc[:, 'lon'])]
		merged = pd.merge(clustered, trajectories_frame, left_index=True, right_index=True, how='outer',
		                  suffixes=['', 'y'])
		merged = merged[['lat', 'lon', 'geometry', 'labels', 'start', 'end']]
		merged[merged['labels'] == -1] = None
		trajectories_frame = TrajectoriesFrame(merged)
		return trajectories_frame


class LayerAggregator():
	"""
	A class for movement trajectories spatial aggregation. This one uses external data (spatial layer) to aggregate data.
	"""

	def __init__(self, layer, kwargs={}):
		"""
		Class initalisation.

		Args:
		layer: An external, spatial layer which will be used for aggregation
		"""
		self._layer = gpd.GeoDataFrame.from_file(layer, **kwargs)

	@property
	def layer(self):
		return self._layer

	@layer.setter
	def layer(self, new):
		self._layer = new

	def aggregate(self, trajectories_frame, dropoverwirte=True):
		"""
		Aggregates data to given spatial layer. Overwrites geometry.

		Args:
			trajectories_frame: TrajectoriesFrame object class
			dropoverwirte: If true, points that were not assigned to any locationare removed.

		Returns:
			An aggregated TrajectoriesFrame instance
		"""
		self._layer = self.layer.to_crs(trajectories_frame.crs)
		aggregated_frame = gpd.tools.sjoin(trajectories_frame.reset_index(), self.layer, how='left')
		centroid_layer = self.layer['geometry'].centroid
		aggregated_frame = aggregated_frame[["user_id", "datetime", 'index_right']]
		aggregated_frame = pd.merge(aggregated_frame, pd.DataFrame(centroid_layer), left_on='index_right',
		                            right_index=True)
		if dropoverwirte:
			trajectories_frame = trajectories_frame.join(aggregated_frame.set_index(["user_id", "datetime"]), how='inner')
		else:
			trajectories_frame = trajectories_frame.join(aggregated_frame.set_index(["user_id", "datetime"]))
		trajectories_frame['geometry'] = trajectories_frame.iloc[:, -1]
		trajectories_frame = trajectories_frame[trajectories_frame.columns[:-1]]
		trajectories_frame['lon'] = trajectories_frame.geometry.x
		trajectories_frame['lat'] = trajectories_frame.geometry.y
		trajectories_frame = TrajectoriesFrame(trajectories_frame)
		return trajectories_frame



import numpy as np
import tqdm
from humobi.misc.utils import get_diags, normalize_chain, _equally_sparse_match


class Sparse(object):
	"""
	Sparse predictor
	"""

	def __init__(self, sequence):
		self._sequence = sequence
		self.model = self.build()

	def build(self):
		scanthrough = {}
		for n in tqdm.tqdm(range(1, len(self._sequence)*2),total=len(self._sequence)*2-1):
			cur_id = len(self._sequence) - n
			if cur_id > 0:
				lookback = self._sequence[cur_id:]
				search_space = self._sequence[:cur_id]
			elif cur_id < 0:
				lookback = self._sequence[:cur_id]
				search_space = self._sequence[cur_id:]
			out = _equally_sparse_match(lookback, search_space)
			if out:
				for ids, candidate in out:
					if candidate in scanthrough.keys():
						scanthrough[candidate].append(ids)
					else:
						scanthrough[candidate] = [ids]
		for k, v in scanthrough.items():
			# v = map(tuple, v)
			# v = list(set(v))
			lists, counts = np.unique(scanthrough[k],return_counts=True)
			if lists.dtype != object:
				if len(lists) > 2:
					lists = [[tuple(lists[[0,2]])], [tuple(lists[[1,2]])]]
					counts = counts[:2]
				else:
					lists = [tuple(lists)]
					counts = counts[0]
			# scanthrough[k] = v
			scanthrough[k] = (lists,counts)
		return scanthrough

	def predict(self, context):
		matches = {}
		prob_dict = {}
		for candidate, ids in self.model.items():
			cnts = ids[1]
			ids = ids[0]
			if not isinstance(ids[0], list):
				ids = [ids]
				cnts = [cnts]
			for cases, each_count in zip(ids,cnts):
				cases = [(x,y) for x,y in cases if abs(x) <= len(context)]
				partial_match = (context[[int(x[0]) for x in cases]] == np.array([x[1] for x in cases]))
				if partial_match.any():
					if candidate in matches.keys():
						matches[candidate].append((cases, partial_match,each_count)) #added count
					else:
						matches[candidate] = [(cases, partial_match,each_count)] #added count
		for candidate, match in matches.items():
			match_fil = [np.array(x[0])[x[1]] for x in match]
			weights = [x**2 for x in range(len(self._sequence))]
			recency = [abs(1 / x[:, 0].sum()) for x in match_fil]
			# recency = [((len(self._sequence)+x[:, 0])/len(self._sequence)) for x in match_fil]
			recency = [max([weights[int(z)]*z for z in (len(self._sequence)+x[:, 0])]) for x in match_fil]
			all_counts = [x[2] for x in match]
			recency = [x * y for x, y in zip(recency, all_counts)]
			prob_dict[candidate] = [a * b for a, b in
			                        zip([x[:, 1].shape[0] / len(y) for x, y in zip(match_fil, match)], recency)]
		prob_dict = {k: sum((x)) for k, x in prob_dict.items()}
		try:
			normalize_chain(prob_dict)
			SMC = max(prob_dict, key=prob_dict.get)
		except ZeroDivisionError:
			SMC = np.argmax(np.unique(context, return_counts=True)[1])
		return SMC



import pandas as pd
import numpy as np


def top_time(ind=None, gs=None):
	"""
	Selects the location (by coordinates) which was visited for the longest period during given time interval

	Args:
		ind: user id
		gs: GeoDataFrame from groupby execution containing all the data in the given time interval

	Returns:
		user id (if given) and the data for the longest visited location
	"""
	aggregated = []
	for tstamp, g in gs:  # for each record in the GeoDataFrame
		if len(g) > 1:  # if there is more than one record
			diff_places = (g['geometry'].shift(-1) != g['geometry']).iloc[:-1]  # checks when coordinates change
			if diff_places.any():  # if there is change in locations
				g_res = g.reset_index()  # drop index
				diffs = g_res.shift(-1)['datetime'] - g_res['datetime']  # find time differences (spent in location)
				joined_dfs = g_res.join(diffs, rsuffix='a')  # add them to locations
				joined_dfs['geometry'] = g_res['geometry'].astype(str)  # copy geometry as string
				point_max = joined_dfs.groupby('geometry')['datetimea'].sum().idxmax()  # grouping locations find the longest time sum
				selected = g[g['geometry'].astype(str) == point_max]  # select the location with the highest total time
			else:
				selected = g  # if one location visited - copy GeoDataFrame
		else:
			selected = g
		aggregated.append(selected)
	if ind is None:
		return pd.concat(aggregated)
	else:
		return ind, pd.concat(aggregated)


def mode_geoseries(ind, gs):
	"""
	Calculates mode for GeoSeries

	Args:
		ind: identifier
		gs: GeoSeries

	Returns:
		identifier and a mode for GeoSeries
	"""
	aggregated = []
	for g in gs:
		if g[1].empty:
			aggregated.append(None)
		else:
			selected = g[1].mode()
			selected = selected.set_index(g[1].index)
			aggregated.append(selected)
	return ind, pd.concat(aggregated)


def rowwise_average(gs, row_count=None):
	"""
	Calculates an average for each row in each group - rowwise.

	Args:
		gs: GeoSeries
		row_count: defines how much rows should be considered

	Returns:
		averaged GeoSeries rowwise
	"""
	if row_count is None:
		row_count = gs.groupby(level=0).size().max()
	return pd.Series([gs.groupby(level=0).nth(n).mean() for n in range(row_count)])


def groupwise_average(gs):
	"""
	Calculates an average from each group of GeoSeries

	Args:
		gs: GeoSeries

	Returns:
		averaged GeoSeries
	"""
	return gs.groupby(level=0).mean()


def groupwise_normalise(gs):
	"""
	Normalises each group of GeoSeries

	Args:
		gs: GeoSeries

	Returns:
		normalised GeoSeries
	"""
	return gs.groupby(level=0).apply(lambda x: x / x.sum())


def groupwise_expansion(gs):
	"""
	Calculates expanding mean for each group of GeoSeries

	Args:
		gs: GeoSeries

	Returns:
		averaged GeoSeries
	"""
	return gs.groupby(level=0).expanding().mean()


def convert_to_distribution(gs, bin_size=None, num_of_classes=None):
	"""
	Converts statistics from Series into a distribution

	Args:
		gs: GeoSeries with a statistic
		bin_size: size of a bin for histogram (if used, num_of_classes cannot be determined)
		num_of_classes: number of groups in histogram (if used, bin_size cannot be determined)

	Returns:
		A Series with distribution of desired parameters. Bins are indicated as index.
	"""
	lower = gs.min()
	upper = gs.max()
	if num_of_classes is None:
		gs = gs.groupby(pd.cut(gs, np.arange(lower, upper, bin_size))).count()
	if bin_size is None:
		gs = gs.groupby(pd.cut(gs, np.linspace(lower, upper, num_of_classes+1))).count()
	return gs/gs.sum()


def normalize(gs, name):
	"""
	Normalises a column of geoseries.

	Args:
		gs: GeoSeries
		name: Column to normalise

	Retruns:
		GeoSeries with given column normalised
	"""
	gs_column = gs.loc[:, name]
	gs_column = gs_column / gs_column.sum()
	if gs_column.sum() != 1:
		offset = 1 - gs_column.sum()
		max_ind = gs_column.idxmax()
		gs_column.loc[max_ind] += offset
	gs.loc[:, name] = gs_column
	return gs


def total_normalise(gs):
	"""
	Performs complete normalisation of GeoSeries

	Args:
		gs: GeoSeries

	Returns:
		normalised GeoSeries
	"""
	return gs / gs.sum()


def start_end(trajectories_frame):
	"""
	Compresses stops in TrajectoriesFrame by adding start and end of visits in locations

	Args:
		trajectories_frame: TrajectoriesFrame object class

	Returns:
		compressed TrajectoriesFrame
	"""
	to_concat = []
	if 'date' not in trajectories_frame.columns:
		trajectories_frame['date'] = trajectories_frame.index.get_level_values(1)
	for gs in trajectories_frame.groupby(level=0):
		firsts = gs[1][gs[1]['geometry'].shift() != gs[1]['geometry']]
		lasts = gs[1][gs[1]['geometry'].shift(-1) != gs[1]['geometry']]
		firsts.loc[:, 'start'] = firsts['date']
		lasts = lasts.set_index(firsts.index)
		firsts.loc[:, 'end'] = lasts['date']
		firsts = firsts[firsts['start'] != firsts['end']]
		to_concat.append(firsts)
	return pd.concat(to_concat)



import pandas as pd
import sys, statistics
import numpy as np
from sklearn.cluster import DBSCAN, KMeans
from sklearn.metrics import silhouette_score, calinski_harabasz_score
sys.path.append("..")
from humobi.models.spatial_tools.misc import rank_freq, normalize_array
from scipy import stats
from collections import Counter
WEIGHT = False
import geopandas as gpd
import os


def cluster_trajectories(trajectories_frame, length = 24, quantity = 2, weights = True, clust_alg = DBSCAN(),
                         aux_cols = None):
	"""
	Extracts circadian rhythms and clusters users by them.
	:param trajectories_frame: TrajectoriesFrame class object
	:param length: The lenght of circadian rhythms to extract
	:param quantity: The number of top most important locations to consider
	:param weights: Whether the algorithm should calculate the weight for each of the locations rather than choose the
	most often visited
	:param clust_alg: Clustering algorithm to make clusterization
	:return: Clustered circadian rhythms, association of users to clusters, the ratio of users in clusters
	"""
	if aux_cols:
		weights = True
		warnings.warn("Warning: when aux cols are passed, the model is enforced to represent abtract trajectory as a set probabilities")
	if len(aux_cols) >1:
		unique_combs = [z for z in product(*[list(pd.unique(trajectories_frame[col].dropna())) for col in aux_cols])]
	else:
		unique_combs = [pd.unique(trajectories_frame[col].dropna()) for col in aux_cols][0]
	top_places = rank_freq(trajectories_frame, quantity)
	abstract_traj = {}
	if length <= 24:
		trajectories_frame['hod'] = trajectories_frame.index.get_level_values(1).hour
	grouped = trajectories_frame.groupby(level=0)
	for uid, vals in grouped:
		extract = vals.groupby(['labels',*aux_cols, 'hod']).count().iloc[:, 0].unstack().fillna(0)
		sig_places = []
		for n in range(quantity):
			sig_place = top_places.loc[uid][n]
			if sig_place is not None:
				sig_place_label = int(vals[vals['geometry'] == sig_place]['labels'].iloc[0])
				extraction_combs = pd.concat([pd.DataFrame(index=unique_combs),extract.loc[sig_place_label,:]],axis=1).fillna(0)
				sig_places.append(extraction_combs)
			else:
				sig_places.append(np.zeros((len(unique_combs),length)))
		stacked = np.vstack(sig_places)
		others = extract.sum(0) - stacked.sum(0)
		stacked = np.vstack((stacked,others))
		if weights:
			abstract_traj[uid] = stacked/stacked.sum(axis=0) #Circadian rhythm
		else:
			abstract_traj[uid] = np.argmax(stacked,axis=0) #most commonly visited place at given time (0-HOME, 1 - WORK, 2 - OTHER for q=2)
	reshaped = np.concatenate([x.reshape(1, -1) for x in abstract_traj.values()], 0) #slices to n hours strips and sets them horizontally in a matrix
	cdist = np.zeros((reshaped.shape[0],reshaped.shape[0]))
	for n in range(reshaped.shape[0]):
		for m in range(reshaped.shape[0]):
			cdist[m,n] = stats.wasserstein_distance(reshaped[n,:],reshaped[m,:])
	tries = {}
	for z in range(1,5000): #find optimal clustering
		eps = z/1000
		try:
			clust_alg = DBSCAN(metric='precomputed',min_samples=4,eps=eps)
			clust_alg.fit(cdist)
			sh = silhouette_score(cdist,clust_alg.labels_,metric='precomputed')
			tries[eps] = sh
		except:
			pass
	eps = max(tries, key=tries.get)
	clust_alg = DBSCAN(metric='precomputed', min_samples=4, eps=eps)
	clust_alg.fit(cdist)
	labels = clust_alg.labels_
	cluster_association = {k: v for k, v in zip(abstract_traj.keys(), clust_alg.labels_)}
	abstract_collection = pd.DataFrame(reshaped, index=labels)
	cluster_share = Counter([x for x in cluster_association.values() if x != -1])
	cluster_share = {k: v / sum(cluster_share.values()) for k, v in cluster_share.items()}
	return abstract_collection, cluster_association, cluster_share


def circadian_rhythm_extraction(circadian_collection):
	"""
	Calculates average circadian rhythm.
	:param circadian_collection: contains circadian rhythms
	:return: a dict with clusters as keys and avarage circadian rhythms as values.
	"""
	cluster_indicies = set(circadian_collection.index.difference(set([-1])))
	circadian_rhythm_grouped = {k:v for k,v in zip(cluster_indicies,[circadian_collection.loc[cluster]
																	 for cluster in cluster_indicies])}
	frequency = {}
	for k, v in circadian_rhythm_grouped.items():
		frequency[k] = dict(zip(np.unique(v, return_counts = True)[0],
						 np.unique(v, return_counts = True)[1]))
	if WEIGHT:
		extracted = {k:v.mean() for k,v in circadian_rhythm_grouped.items()}
	elif not WEIGHT:
		extracted = {k:v.mode() for k,v in circadian_rhythm_grouped.items()}
		for k,v in extracted.items():
			if len(v) > 1:
				most_freq = max(frequency[k].items(), key = lambda x: x[1])[0]
				v.iloc[0][np.where(np.array(np.isnan(v.iloc[1])) == False)[0][0]] = most_freq
			extracted[k] = v.iloc[:1]
	return extracted
